""" SUBCOMMAND:  ``ksconf unarchive <tarball>``

Usage example:

.. code-block:: sh

    ksconf unarchive splunk-add-on-for-amazon-web-services_111.tgz

"""

from __future__ import absolute_import, print_function, unicode_literals

import os
import re
from pathlib import Path
from subprocess import list2cmdline

from ksconf.app import get_facts_manifest_from_archive
from ksconf.app.facts import AppFacts
from ksconf.app.manifest import AppArchiveContentError, AppArchiveError
from ksconf.archive import extract_archive, gen_arch_file_remapper
from ksconf.command import KsconfCmd, dedent
from ksconf.conf.parser import PARSECONF_LOOSE, ConfParserException, parse_conf
from ksconf.consts import (EXIT_CODE_BAD_ARCHIVE_FILE, EXIT_CODE_BAD_ARGS,
                           EXIT_CODE_FAILED_SAFETY_CHECK, EXIT_CODE_GIT_FAILURE,
                           is_debug)
from ksconf.filter import create_filtered_list
from ksconf.util.compare import cmp_sets
from ksconf.util.completers import DirectoriesCompleter, FilesCompleter
from ksconf.util.file import dir_exists, file_hash, relwalk
from ksconf.vc.git import (git_cmd, git_cmd_iterable, git_is_clean,
                           git_is_working_tree, git_ls_files, git_status_ui,
                           git_version)

allowed_extensions = ("*.tgz", "*.tar.gz", "*.spl", "*.zip")


# XXX:  Add a git status --ignored --porcelain APPNAME check to list out files/dirs that are excluded...

# XXX:  Update code base to use ksconf.layer, as was done for the 'ksconf combine' command.


DEFAULT_DIR = "default"


class UnarchiveCmd(KsconfCmd):
    help = "Install or upgrade an existing app in a git-friendly and safe way"
    description = dedent("""
    Install or overwrite an existing app in a git-friendly way.
    If the app already exists, steps will be taken to upgrade it safely.

    The ``default`` folder can be redirected to another path (i.e., ``default.d/10-upstream`` or
    other desirable path if you're using the ``ksconf combine`` tool to manage extra layers).
    """)
    format = "manual"
    maturity = "beta"

    def register_args(self, parser):
        parser.add_argument("tarball", metavar="SPL",
                            help="The path to the archive to install."
                            ).completer = FilesCompleter(allowednames=allowed_extensions)
        parser.add_argument("--dest", metavar="DIR", default=".", help=dedent("""\
            Set the destination path where the archive will be extracted.
            By default, the current directory is used.  Sane values include: etc/apps,
            etc/deployment-apps, and so on.""")
                            ).completer = DirectoriesCompleter()
        parser.add_argument("--app-name", metavar="NAME", default=None, help=dedent("""\
            The app name to use when expanding the archive.
            By default, the app name is taken from the archive as the top-level path included
            in the archive (by convention).
            """))
        parser.add_argument("--default-dir", default=DEFAULT_DIR, metavar="DIR", help=dedent("""\
            Name of the directory where the default contents will be stored.
            This is a useful feature for apps that use a dynamic default directory
            that's created and managed by the 'combine' mode.""")
                            ).completer = DirectoriesCompleter()
        parser.add_argument("--exclude", "-e", action="append", default=[], help=dedent("""\
            Add a file pattern to exclude from extraction.
            Splunk's pseudo-glob patterns are supported here.
            ``*`` for any non-directory match,
            ``...`` for ANY (including directories),
            and ``?`` for a single character."""))
        parser.add_argument("--keep", "-k", action="append", default=[],
                            help=dedent("""\
            Specify a pattern for files to preserve during an upgrade.
            Repeat this argument to keep multiple patterns."""))
        parser.add_argument("--allow-local", default=False, action="store_true", help=dedent("""\
            Allow local/* and local.meta files to be extracted from the archive.
            """))
        parser.add_argument("--git-sanity-check",
                            choices=["off", "changed", "untracked", "ignored"],
                            default="untracked", help=dedent("""\
            By default, 'git status' is run on the destination folder to detect working tree or
            index modifications before the unarchive process start.

            Sanity check choices go from least restrictive to most thorough:

            'off' prevents all safety checks.
            'changed' aborts only upon local modifications to files tracked by git.
            'untracked' (the default) looks for changed and untracked files.
            'ignored' aborts is (any) local changes, untracked, or ignored files are found.
            """))
        parser.add_argument("--git-mode", default="stage",
                            choices=["nochange", "stage", "commit"], help=dedent("""\
            Set the desired level of git integration.
            The default mode is *stage*, where new, updated, or removed files are automatically
            handled for you.

            To prevent any ``git add`` or ``git rm`` commands from being run, pick the
            'nochange' mode.
            """))
        parser.add_argument("--no-edit",
                            action="store_true", default=False, help=dedent("""\
            Tell git to skip opening your editor on commit.
            By default, you will be prompted to review/edit the commit message.
            (Git Tip:  Delete the content of the default message to abort the commit.)"""))
        parser.add_argument("--git-commit-args", "-G", default=[], action="append",
                            help="Extra arguments to pass to 'git'")

    def run(self, args):
        """ Install / upgrade a Splunk app from an archive file """
        # Handle ignored files by preserving them as much as possible.
        # Add --dry-run mode?  j/k - that's what git is for!
        DEBUG = is_debug()

        tarball = Path(args.tarball)
        if not tarball.is_file():
            self.stderr.write(f"No such file or directory {tarball}\n")
            return EXIT_CODE_FAILED_SAFETY_CHECK

        dest = Path(args.dest)
        if not dest.is_dir():
            self.stderr.write(f"Destination directory does not exist: {dest}\n")
            return EXIT_CODE_FAILED_SAFETY_CHECK

        f_hash = file_hash(tarball)
        self.stdout.write(f"Inspecting archive:               {tarball}\n")

        new_app_name = args.app_name

        if new_app_name and "/" in new_app_name:
            if new_app_name.endswith("/"):
                new_app_name = new_app_name.rstrip("/")
            else:
                self.stderr.write(f"Invalid app name.  Please remove '/' from app name {new_app_name}")
                return EXIT_CODE_BAD_ARGS

        # ARCHIVE PRE-CHECKS:  Archive must contain only one app, no weird paths, ...
        try:
            app_facts, app_manifest = get_facts_manifest_from_archive(
                tarball, calculate_hash=False, check_paths=True)
        except AppArchiveError as e:
            self.stderr.write(f"Failed to extract content from {tarball}\n{e}")
            return EXIT_CODE_BAD_ARCHIVE_FILE
        except AppArchiveContentError as e:
            self.stderr.write(f"The 'unarchive' command does not support the {tarball} archive.\n{e}")
            return EXIT_CODE_FAILED_SAFETY_CHECK

        local_files = set(app_manifest.find_local())
        app_name = app_facts.name

        if local_files:
            self.stderr.write(f"Local {len(local_files)} files found in the archive.  ")
            if args.allow_local:
                self.stderr.write("Keeping these due to the '--allow-local' flag\n")
            else:
                self.stderr.write("Excluding local files by default.  "
                                  "Use '--allow-local' to override.")

        if not new_app_name and True:  # if not --no-app-name-fixes
            if app_name.endswith("-master"):
                self.stdout.write("Automatically dropping '-master' from the app name.  "
                                  "This is often the result of a github export.\n")
                new_app_name = app_name[:-7]
            mo = re.search(r"(.*)-\d+\.[\d.-]+$", app_name)
            if mo:
                new_app_name = mo.group(1)
                self.stdout.write("Automatically removing the version suffix from the app name.  "
                                  f"'{app_name}' will be extracted as '{new_app_name}'\n")

        app_basename = new_app_name or app_name
        dest_app: Path = dest / app_basename
        self.stdout.write(f"Inspecting destination folder:    {dest_app.absolute()}\n")

        # FEEDBACK TO THE USER:   UPGRADE VS INSTALL, GIT?, APP RENAME, ...
        app_name_msg = app_name

        git_ver = git_version()
        if git_ver is None:
            vc_msg = "without version control support (git not present)"
            is_git = False
        else:
            vc_msg = "without version control support"

        existing_app = None

        if dest_app.is_dir():
            mode = "upgrade"
            if git_ver:
                is_git = git_is_working_tree(dest_app)
            try:
                # Ignoring the 'local' entries since distributed apps shouldn't contain local
                old_app_conf_file = os.path.join(dest_app, args.default_dir, "app.conf")
                old_app_conf = parse_conf(old_app_conf_file, profile=PARSECONF_LOOSE)
                existing_app = AppFacts.from_conf(app_basename, old_app_conf)
                del old_app_conf
            except (ConfParserException, FileNotFoundError):
                self.stderr.write("Unable to read app.conf from existing install.\n")
                # Assume upgrade form unknown version
        else:
            mode = "install"
            if git_ver:
                is_git = git_is_working_tree(args.dest)
        if is_git:
            vc_msg = "with git support"

        if new_app_name:
            if new_app_name == app_name:
                # No rename detected (user provided 'app_name' matches the default app name)
                new_app_name = None
            else:
                app_name_msg = f"{new_app_name} (renamed from {app_name})"

        def show_pkg_info(facts: AppFacts, label):
            self.stdout.write(f"{label} packaging info:    "
                              f"'{facts.label or 'Unknown'}' by "
                              f"{facts.author or 'Unknown'} "
                              f"(version {facts.author or 'Unknown'})\n")

        if existing_app:
            show_pkg_info(existing_app, " Installed app")
        if app_facts:
            show_pkg_info(app_facts, "   Tarball app")

        self.stdout.write(f"About to {mode} the {app_name_msg} app {vc_msg}.\n")

        existing_files = set()
        if mode == "upgrade":
            if is_git:
                existing_files.update(git_ls_files(dest_app))
                if not existing_files:
                    self.stderr.write("App is in a git repository but no files have been staged "
                                      f"or committed.  Either commit or remove '{dest_app}' and try again.")
                    return EXIT_CODE_FAILED_SAFETY_CHECK
                if args.git_sanity_check == "off":
                    self.stdout.write("The 'git status' safety checks have been disabled via CLI"
                                      "argument.  Skipping.\n")
                else:
                    d = {
                        #        untracked, ignored
                        "changed": (False, False),
                        "untracked": (True, False),
                        "ignored": (True, True)
                    }
                    is_clean = git_is_clean(dest_app, *d[args.git_sanity_check])
                    del d
                    if is_clean:
                        self.stdout.write("Git folder is clean.  "
                                          "Okay to proceed with the upgrade.\n")
                    else:
                        self.stderr.write("Unable to move forward without a clean working tree.\n"
                                          "Clean up and try again.  "
                                          "Modifications are listed below.\n\n")
                        self.stderr.flush()
                        if args.git_sanity_check == "changed":
                            git_status_ui(dest_app, "--untracked-files=no")
                        elif args.git_sanity_check == "ignored":
                            git_status_ui(dest_app, "--ignored")
                        else:
                            git_status_ui(dest_app)
                        return EXIT_CODE_FAILED_SAFETY_CHECK
            else:
                for (root, _, filenames) in relwalk(dest_app):
                    for fn in filenames:
                        existing_files.add(os.path.join(root, fn))
            self.stdout.write(f"Before upgrade.  App has {len(existing_files)} files\n")
        elif is_git:
            self.stdout.write("Git clean check skipped.  Not needed for a fresh app install.\n")

        def fixup_pattern_bw(patterns, prefix=None):
            modified = []
            for pattern in patterns:
                if pattern.startswith("./"):
                    if prefix:
                        pattern = f"{prefix}/{pattern[2:]}"
                    else:
                        pattern = pattern[2:]
                    modified.append(pattern)
                # If a pattern like 'tags.conf' or '*.bak' is provided, use basename match (any dir)
                elif "/" not in pattern:
                    modified.append("(^|.../)" + pattern)
                else:
                    modified.append(pattern)
            return modified

        # PREP ARCHIVE EXTRACTION
        installed_files = set()
        excludes = list(args.exclude)
        if not args.allow_local:
            for pattern in local_files:
                excludes.append("./" + pattern)
        excludes = fixup_pattern_bw(excludes, app_basename)
        self.stderr.write(f"Extraction exclude patterns:  {excludes!r}\n")
        exclude_filter = create_filtered_list("splunk", default=False)
        exclude_filter.feedall(excludes)

        # Calculate path rewrite operations
        path_rewrites = []
        files_iter = extract_archive(args.tarball)
        if args.default_dir != DEFAULT_DIR:
            rep = rf"\1/{ args.default_dir.strip('/') }/"
            path_rewrites.append((re.compile(rf"^(/?[^/]+)/{DEFAULT_DIR}/"), rep))
            del rep
        if new_app_name:
            regex = re.compile(r'^([^/]+)(?=/)')
            path_rewrites.append((regex, new_app_name))
            del regex
        if path_rewrites:
            files_iter = gen_arch_file_remapper(files_iter, path_rewrites)
        del path_rewrites

        # Filer out "removed" files; and let us keep some based on a keep-allowlist
        self.stdout.write("Extracting app now...\n")
        for gaf in files_iter:
            if exclude_filter.match(gaf.path):
                self.stdout.write(f"Skipping [blocklist] {gaf.path}\n")
                continue
            if not is_git or args.git_mode in ("nochange", "stage"):
                self.stdout.write(f"{gaf.path:60s} {gaf.mode:o} {gaf.size:-6d}\n")
            installed_files.add(gaf.path.split("/", 1)[1])
            full_path = os.path.join(args.dest, gaf.path)
            dir_exists(os.path.dirname(full_path))
            with open(full_path, "wb") as fp:
                fp.write(gaf.payload)
            os.chmod(full_path, gaf.mode)
            del fp, full_path

        files_new, files_upd, files_del = cmp_sets(installed_files, existing_files)

        if DEBUG:
            def dbg_fmt(l):
                return "\n\t".join(sorted(l))

            print(f"New:      \n\t{dbg_fmt(files_new)}")
            print(f"Existing: \n\t{dbg_fmt(files_upd)}")
            print(f"Removed:  \n\t{dbg_fmt(files_del)}")

        self.stdout.write(f"Extracted {len(installed_files)} files:  "
                          f"{len(files_new)} new, {len(files_upd)} existing, "
                          f"and {len(files_del)} removed\n")

        # Filer out "removed" files; and let us keep some based on a keep-allowlist:  This should
        # include things like local, ".gitignore", ".gitattributes" and so on

        keep_list = [".git*"]
        keep_list.extend(args.keep)
        if not args.allow_local:
            keep_list += ["local/...", "local.meta"]
        keep_list = fixup_pattern_bw(keep_list)
        self.stderr.write(f"Keep file patterns:  {keep_list!r}\n")

        keep_filter = create_filtered_list("splunk", default=False)
        keep_filter.feedall(keep_list)

        files_to_delete = []
        files_to_keep = []
        for fn in files_del:
            if keep_filter.match(fn):
                # How to handle a keep of "default.d/..." when we DO want to cleanup the default
                # redirect folder of "default.d/10-upstream"?
                # This may be an academic question since most apps will continue to send
                # an ever increasing list of default files (to mask out old/unused ones)
                self.stdout.write(f"Keeping {fn}\n")
                files_to_keep.append(fn)
            else:
                files_to_delete.append(fn)
        if files_to_keep:
            self.stdout.write(f"Keeping {len(files_to_keep)} of {len(files_del)} "
                              "files marked for deletion due to allow list.\n")
        git_rm_queue = []

        if files_to_delete:
            self.stdout.write("Removing files not present in the upgraded version of the app.\n")
        for fn in files_to_delete:
            path = os.path.join(dest_app, fn)
            if is_git and args.git_mode in ("stage", "commit"):
                self.stdout.write(f"git rm -f {path}\n")
                git_rm_queue.append(fn)
            else:
                self.stdout.write(f"rm -f {path}\n")
                os.unlink(path)

        if git_rm_queue:
            # Run 'git rm file1 file2 file3 ..." (using an xargs like mechanism)
            git_cmd_iterable(["rm"], git_rm_queue, cwd=dest_app)
        del git_rm_queue

        if is_git:
            if args.git_mode in ("stage", "commit"):
                git_cmd(["add", "--all", dest_app.name], cwd=dest_app.parent)
                # self.stdout.write(f"git add {os.path.basename(dest_app)}\n")
            '''
            else:
                self.stdout.write(f"git add {dest_app}\n")
            '''

            # Is there anything to stage/commit?
            if git_is_clean(dest_app.parent, check_untracked=False):
                self.stderr.write(f"No changes detected.  Nothing to {args.git_mode}\n")
                return

            git_commit_app_name = app_facts.label or dest_app.name
            if mode == "install":
                git_commit_message = f"Install {git_commit_app_name}"
                if app_facts.version:
                    git_commit_message += f" version {app_facts.version}"
            else:
                # Todo:  Specify Upgrade/Downgrade/Refresh
                git_commit_message = f"Upgrade {git_commit_app_name}"
                if existing_app.version and app_facts.version:
                    git_commit_message += f" version {app_facts.version} (was {existing_app.version})"
                elif app_facts.version:
                    git_commit_message += f" to version {app_facts.version}"
            # Could possibly include some CLI arg details, like what file patterns were excluded
            git_commit_message += f"\n\nSHA256 {f_hash} {tarball.name}\n\nSplunk-App-managed-by: ksconf"
            git_commit_cmd = ["commit", dest_app.name, "-m", git_commit_message]

            if not args.no_edit:
                git_commit_cmd.append("--edit")

            git_commit_cmd.extend(args.git_commit_args)

            if args.git_mode == "commit":
                capture_std = True if args.no_edit else False
                proc = git_cmd(git_commit_cmd, cwd=dest_app.parent, capture_std=capture_std)
                if proc.returncode == 0:
                    self.stderr.write(dedent("""\
                    Your changes have been committed.  Please review before pushing.  If you
                    find any issues, here are some possible solutions:


                    To fix issues in the last commit, edit and add the files to be fixed, then run:

                        git commit --amend

                    To roll back the last commit but KEEP the app upgrade, run:

                        git reset --soft HEAD^1

                    To roll back the last commit and REVERT the app upgrade, run:

                        git reset --hard HEAD^1

                    NOTE:  Make sure you have *no* other uncommitted changes before running 'reset'.
                    """))
                else:
                    self.stderr.write(f"Git commit failed.  Return code {proc.returncode}.  "
                                      f"Git args:  git {list2cmdline(git_commit_cmd)}\n")
                    return EXIT_CODE_GIT_FAILURE
            elif args.git_mode == "stage":
                self.stdout.write("To commit later, use the following\n")
                self.stdout.write("\tgit {}\n".format(
                    list2cmdline(git_commit_cmd).replace("\n", "\\n")))
            # When in 'nochange' mode, no point in even noting these options to the user.



"""
This module contains all the plugin definitions (or hook "specifications") for various customization
or integration points with ksconf.  Not all of these have been fully tested so please let us know if
something is not working as expected, or if additional arguments are needed.

See `ksconf plugins <https://pypi.org/search/?q=ksconf&o=&c=Environment+%3A%3A+Plugins>`__ on pypi
for a list of currently available plugins.
"""

from pathlib import Path
from typing import TYPE_CHECKING, Any

from pluggy import HookspecMarker

from ksconf.consts import PLUGGY_HOOK

if TYPE_CHECKING:
    from argparse import ArgumentParser, Namespace

    from jinja2 import Environment
else:
    Environment = Any
    ArgumentParser = Namespace = Any

# For Python 3.7 (Protocol added in 3.8)
try:
    from typing import Protocol
except ImportError:
    class Protocol:
        pass


class KsconfPluginWarning(Warning):
    pass


hookspec = HookspecMarker(PLUGGY_HOOK)


# Definitions of ksconf defined hooks

# Additional hooks to add someday:
#
#
# ksconf_cli_exit(return_code)
# ksconf_cli_unhandled_exception(exc_info...)
#


class KsconfHookSpecs(Protocol):
    """ Ksconf plugin specifications for all known supported functions.

    Grouping these functions together in a single class allows for type support
    it supports typing.  This adds a level of validation to the code base where
    a hook is invoked via ``plugin_manger.hook.<hook_name>()``.

    If you are implementing one of these hooks, please note that you can simple
    make top-level function, no need to implement a class.
    """
    # All methods are decorated with staticmethod, for equivalence with a plain function.

    @staticmethod
    @hookspec
    def ksconf_cli_init():
        """
        Simple hook that is run before CLI initialization.  This can be use to
        modify the runtime environment.

        This can be used to register additional handlers, such as:

        * :py:func:`ksconf.combine.register_handler` - Add a combination file handler.
          File types are limited to pattern matching.
        * :py:func:`ksconf.layer.register_file_handler` - Add file handlers for
          layer processing for template processing
        """
        ...

    @staticmethod
    @hookspec
    def ksconf_cli_modify_argparse(parser: ArgumentParser, name: str):
        """
        Manipulate argparse rules.  This could be used to add additional CLI
        options for other hook-added features added features

        Note that this hook is called for both the top-level argparse instance
        as well as each subparser.  The :py:obj:`name` argument should be
        inspected to determine if the parse instances is the parent (top-level)
        parser, or some other named subcommands.
        """
        ...

    @staticmethod
    @hookspec
    def ksconf_cli_process_args(args: Namespace):
        """
        Hook to capture all parsed arguments, includes any custom arguments
        added to the CLI via the the :py:func:`ksconf_cli_modify_argparse` hook.
        :py:obj:`args` can be mutated directly, if needed.
        """
        ...

    @staticmethod
    @hookspec
    def post_combine(target: Path, usage: str):
        """
        Trigger a custom action after a layer combining operation.  This is used
        by multiple ksconf subcommands and the API.

        This trigger could be used to modify the file system, trigger external
        operations, track/audit behaviors, and so on.

        When using CLI commands, :py:obj:`usage` should be either "combine" or
        "package" depending on which ksconf command was invoked.
        Direct invocation of :py:class:`~ksconf.combine.LayerCombiner` can pass
        along a custom usage label and avoid impacting CLI, when desirable.

        If your goal is to only trigger an action during the app packaging
        process, also consider the :py:func:`package_pre_archive` hook, which
        may be more appropriate.
        """
        ...

    @staticmethod
    @hookspec
    def package_pre_archive(app_dir: Path, app_name: str):
        """
        Modify, inventory, or test the contents of an app before the final
        packaging commands.  This can be triggered from the ``ksconf package``
        command or via the API.

        During a ``ksconf package`` process, this hook executes right before
        the final archive is created.  All local merging, app version or build
        updates, and so on are completed before this hook is executed.

        From an API perspective, this hook is  called from
        :py:class:`ksconf.package.AppPackager` whenever a content freeze occurs,
        which is typically when
        :py:meth:`~ksconf.package.AppPackager.make_archive` or
        :py:meth:`~ksconf.package.AppPackager.make_manifest` is invoked.
        """

    @staticmethod
    @hookspec
    def modify_jinja_env(env: Environment):
        """
        Modify the Jinja2 environment object.  This can be used to add custom
        filters or tests, for example.

        Invoked by :py:class:`~ksconf.layer.LayerFile_Jinja2` immediately after
        initial Environment creation.  :py:obj:`env` should be mutated in place.
        """
        ...



from __future__ import unicode_literals

from functools import partial, wraps
from typing import Callable

from ksconf.consts import is_debug


def _xargs(iterable, cmd_len=1024):
    fn_len = 0
    buf = []
    iterable = list(iterable)
    while iterable:
        s = iterable.pop(0)
        l = len(s) + 1
        if fn_len + l >= cmd_len:
            yield buf
            buf = []
            fn_len = 0
        buf.append(s)
        fn_len += l
    if buf:
        yield buf


def decorator_with_opt_kwargs(decorator: Callable) -> Callable:
    """
    Make a decorator that can work with or without args.
    Heavily borrowed from:  https://gist.github.com/ramonrosa/402af55633e9b6c273882ac074760426
    Thanks to GitHub user ramonrosa
    """
    @wraps(decorator)
    def decorator_wrapper(*args, **kwargs):
        if len(kwargs) == 0 and len(args) == 1 and callable(args[0]):
            return decorator(args[0])
        if len(args) == 0:
            return partial(decorator, **kwargs)
        raise TypeError(f"{decorator.__name__} expects either a single Callable "
                        "or keyword arguments")
    return decorator_wrapper


def debug_traceback():  # pragma: no cover
    """ If the 'KSCONF_DEBUG' environmental variable is set, then show a stack trace. """
    level = 10
    if is_debug():
        # TODO:  Pop one off the top of the stack to hide THIS function
        import traceback
        traceback.print_exc(level)



from __future__ import absolute_import, unicode_literals

import os
import shutil
import sys
from copy import deepcopy
from typing import List, Optional

from ksconf.command import ConfFileProxy
from ksconf.conf.delta import compare_cfgs, show_diff
from ksconf.conf.parser import (GLOBAL_STANZA, ConfType, _extract_comments,
                                inject_section_comments, parse_conf, write_conf)
from ksconf.consts import SMART_UPDATE, SmartEnum
from ksconf.util.file import relwalk

####################################################################################################
# Merging logic

# TODO: Replace this with "<<DROP_STANZA>>" on ANY key.  Let's use just ONE mechanism for all of
# these merge hints/customizations
STANZA_MAGIC_KEY = "_stanza"
STANZA_OP_DROP = "<<DROP>>"


def _merge_conf_dicts(base, new_layer):
    """ Merge new_layer on top of base.  It's up to the caller to deal with any necessary object
    copying to avoid odd referencing between the base and new_layer"""
    for (section, items) in new_layer.items():
        if STANZA_MAGIC_KEY in items:
            magic_op = items[STANZA_MAGIC_KEY]
            if STANZA_OP_DROP in magic_op:
                # If this section exist in a parent (base), then drop it now
                if section in base:
                    del base[section]
                continue  # pragma: no cover  (peephole optimization)
        if section in base:
            # TODO:  Support other magic here...
            # Rip all the comments out of the new_layer, and prepend them (sequentially) to base
            comments = _extract_comments(items)
            if comments:
                inject_section_comments(base[section], prepend=comments)
            base[section].update(items)
        else:
            # TODO:  Support other magic here too..., though with no parent info
            base[section] = items
    # Nothing to return, base is updated in-place


def merge_conf_dicts(*dicts: ConfType) -> ConfType:
    result = {}
    for d in dicts:
        d = deepcopy(d)
        if not result:
            result = d
        else:
            # Merge each subsequent layer on one at a time
            _merge_conf_dicts(result, d)
    return result


def merge_conf_files(dest: ConfFileProxy,
                     configs: List[ConfFileProxy],
                     dry_run: bool = False,
                     banner_comment: Optional[str] = None) -> SmartEnum:
    # Parse all config files
    cfgs = [conf.data for conf in configs]
    newest_mtime = max(conf.mtime for conf in configs) if configs else None
    # Merge all config files:
    merged_cfg = merge_conf_dicts(*cfgs)
    if banner_comment:
        if not banner_comment.startswith("#"):
            banner_comment = "#" + banner_comment
        inject_section_comments(merged_cfg.setdefault(GLOBAL_STANZA, {}), prepend=[banner_comment])

    # Either show the diff (dry-run mode) or write to the destination file
    if dry_run and dest.is_file():
        if os.path.isfile(dest.name):
            dest_cfg = dest.data
        else:
            dest_cfg = {}
        show_diff(sys.stdout, compare_cfgs(dest_cfg, merged_cfg),
                  headers=(dest.name, dest.name + "-new"))
        return SMART_UPDATE
    return dest.dump(merged_cfg, mtime=newest_mtime)


def merge_update_conf_file(dest: str,
                           sources: List[str],
                           remove_source: bool = False) -> None:
    """ Dest is treated as both the output, and the highest priority source.
    """
    # XXX:  If dest is missing/empty, and only one non-empty source, use file move
    # XXX:  If dest is present and no sources are present/non-empty, no-op
    # XXX:  Should there be a 'preserve_mtime=True' option?  Any reason to skip this?
    remove = []
    confs = []
    mtimes = []

    if os.path.isfile(dest):
        confs.append(parse_conf(dest))
        mtimes.append(os.stat(dest).st_mtime)
    for source in sources:
        if os.path.isfile(source):
            confs.append(parse_conf(source))
            mtimes.append(os.stat(source).st_mtime)
            remove.append(source)
    if confs:
        # Put in correct order.  (Last read has highest priority)
        confs.reverse()
        write_conf(dest, merge_conf_dicts(*confs))
        mtime = max(mtimes)
        os.utime(dest, (mtime, mtime))

    if remove_source:
        for name in remove:
            os.unlink(name)


def merge_update_any_file(dest: str,
                          sources: List[str],
                          remove_source: bool = False) -> None:
    # XXX: Set this up in a shared function somewhere, we have to figure this out multiple places
    remove = []
    if dest.endswith(".conf") or dest.endswith(".meta"):
        # self.output.write(f"Merge   {local} -> {default}\n")
        merge_update_conf_file(dest, sources, remove_source=remove_source)
        # merge_conf(default, local)
    else:
        sources = [s for s in sources if os.path.isfile(s)]
        if os.path.isfile(dest):
            # Keep dest file, cleanup sources if requested
            remove.extend(sources)
        else:
            if sources:
                # The highest priority (first one on the list) always wins
                source = sources.pop(0)
                if remove_source:
                    shutil.move(source, dest)
                    remove.extend(sources)
                else:
                    shutil.copy(source, dest)
    if remove_source:
        for name in remove:
            os.unlink(name)


def merge_app_local(app_folder: str, cleanup: bool = True) -> None:
    """
    Find everything in local, if it has a corresponding file in default, merge.
    This function assumes standard Splunk app path names.
    """
    local_dir = os.path.join(app_folder, "local")
    default_dir = os.path.join(app_folder, "default")
    local_meta = os.path.join(app_folder, "metadata", "local.meta")
    default_meta = os.path.join(app_folder, "metadata", "default.meta")

    # Handle 'local/', recursively, if present  (assume that symlinks have already been handled)
    if os.path.isdir(local_dir):
        for (root, _, files) in relwalk(local_dir):
            default_d = os.path.join(default_dir, root)
            if not os.path.isdir(default_d):
                os.mkdir(default_d)
            for file in files:
                local_file = os.path.join(local_dir, root, file)
                default_file = os.path.join(default_dir, root, file)
                merge_update_any_file(default_file, [local_file], cleanup)

    # Handle metadata, if present
    merge_update_conf_file(default_meta, [local_meta], cleanup)



from __future__ import absolute_import, annotations, unicode_literals

import argparse
import inspect
import re
import sys
from pathlib import Path
from subprocess import Popen
from typing import Callable, List, Optional, TextIO

from ksconf.consts import EXIT_CODE_INTERNAL_ERROR, is_debug

QUIET = -1
NORMAL = 0
VERBOSE = 1


class BuildExternalException(Exception):
    pass


class BuildCacheException(Exception):
    pass


class BuildStep:
    __slots__ = ["build_path", "source_path", "dist_path", "config", "verbosity", "_output"]

    def __init__(self,
                 build: Path,
                 source: Optional[Path] = None,
                 dist: Optional[Path] = None,
                 output: TextIO = sys.stdout):
        self.build_path = build
        self.source_path = source
        self.dist_path = dist
        self.config = {}
        self.verbosity = 0
        self._output = output

    def alternate_path(self, path) -> BuildStep:
        """ Construct a new BuildStep instance with only 'build_path' altered. """
        cls = self.__class__
        instance = cls(path)
        for slot in cls.__slots__:
            if slot != "build_path":
                setattr(instance, slot, getattr(self, slot))
        return instance

    @property
    def is_quiet(self):
        return self.verbosity <= QUIET

    def is_verbose(self):
        return self.verbosity >= VERBOSE

    def get_logger(self, prefix: Optional[str] = None) -> Callable:
        if prefix is None:
            prefix = inspect.currentframe().f_back.f_code.co_name
        elif re.match(r'[\w_]+', prefix):
            prefix = f"[{prefix}] "

        def log(message, *args, **kwargs):
            message = f"{prefix} {message}"
            return self._log(message, *args, **kwargs)
        return log

    def _log(self, message, verbosity=0):
        """ verbosity:  lower=more important,
            -1 = quiet
             0 = default
            +1 verbose
        """
        if int(verbosity) <= self.verbosity:
            self._output.write(message)
            self._output.write("\n")

    def run(self, executable, *args, cwd=None):
        """ Execute an OS-level command regarding the build process.
        The process will run withing the working directory of the build folder.

        :param str executable: Executable to launch for a build step.
        :param str args: Additional argument(s) for the new process.
        :param str cwd:  Optional kw arg to change the working directory.  This
                         defaults to the build folder.
        """
        # XXX: Update the external pip call to detach stdout / stderr if self.is_quiet
        args = [executable] + [str(s) for s in args]
        cwd = cwd or str(self.build_path)
        exec_info = " ".join(str(s) for s in args)
        self._log(f"EXEC:  {exec_info}  cwd={cwd}", VERBOSE)
        process = Popen(args, cwd=cwd)
        process.wait()
        if process.returncode != 0:
            raise BuildExternalException(f"Exit code of {process.returncode} "
                                         f"while executing {executable}")

    def run_ksconf(self, *args, cwd=None):
        """ Execute 'ksconf' command in the build folder.
        Currently this runs as a separate process, but in the future is may be
        optimized to run from within the same python process.  This is an
        implementation detail the caller shouldn't care about.

        :param str args: Additional argument(s) for the ksconf command.
        :param str cwd:  Optional kw arg to change the working directory.  This
                         defaults to the build folder.
        """
        # Historically '-m ksconf' was used, but here safely skip python version check
        return self.run(sys.executable, "-m", "ksconf.cli", *args, cwd=cwd)


from ksconf.builder.core import BuildManager  # noqa


def default_cli(build_manager: BuildManager,
                build_funct: Callable,
                argparse_parents: List[argparse.ArgumentParser] = ()):
    """
    This is the function you stick in the:  ``if __name__ == '__main__'`` section of your code :-)

    Pass in a BuildManager instance, and a callback function.  The callback function must accept
    (steps, args).  If you have need for custom arguments, you can add them to your own
    ArgumentParser instance and pass them to the argparse_parents keyword argument, and then handle
    additional 'args' passed into the callback function.
    """
    parser = argparse.ArgumentParser(parents=argparse_parents)
    parser.add_argument("--verbose", "-v", action="count", default=0)
    parser.add_argument("--quiet", "-q", action="count", default=0)
    parser.add_argument("--build", metavar="DIR", default="build",
                        help="Set build folder destination")
    parser.add_argument("--no-cache", action="store_true", default=False,
                        help="Disable caching")
    parser.add_argument("--taint-cache",
                        action="store_true", default=False)
    args = parser.parse_args()

    verbosity = args.verbose - args.quiet
    if not build_manager.is_folders_set():
        build_manager.set_folders(".", args.build)
    if args.no_cache:
        build_manager.disable_cache()
    if args.taint_cache:
        build_manager.taint_cache()
    step = build_manager.get_build_step()
    step.verbosity = verbosity

    try:
        return build_funct(step, args)
    except Exception as e:
        if is_debug():
            # XXX: instead of re-raising; write out traceback (with this final frame removed)?
            # Allow stack track to be dumped to screen for developer review/debugging
            raise
        sys.stderr.write(f"Unhandled exception in build process:  {e}\n")
        sys.exit(EXIT_CODE_INTERNAL_ERROR)



"""
Silly simple Python version compatibility items
"""

import sys

# Since 'list' in Python 3.7 doesn't support __class_getitem__ (See PEP 560)
# We could just always import from typing, but that appears to cause some issues
# with type hinting, and this appears to work correctly.  Most of this appears
# to have changed in Python 3.8 and 3.9 (differs by type. hence the test-and-see approach)

''' # Alternate, needs more testing; not sure about the assignment part
# Another option would be to just assign directly into the typing module, what could go wrong?

from typing import Dict, List, Set, Tuple
for t in ("dict", "list", "set", "tuple"):
    try:
        T = t.capitalize()
        eval(f"{t}[str]")
        globals()[T] = globals[t]
    except TypeError:
        pass
'''

if sys.version_info < (3, 9):
    from typing import Dict, List, Set, Tuple
else:
    Dict = dict
    List = list
    Set = set
    Tuple = tuple


try:
    # Python 3.9 and later
    from functools import cache
except ImportError:
    from functools import lru_cache
    cache = lru_cache(maxsize=None)

del sys

__all__ = [
    "Dict",
    "List",
    "Set",
    "Tuple",
    "cache",
]



from __future__ import annotations

import json
from datetime import datetime, timedelta
from pathlib import Path, PurePath
from shutil import copy2, rmtree
from typing import List

from ksconf.builder import BuildCacheException
from ksconf.util.file import file_hash

# Finger print functions for a FileSet operations


def fingerprint_hash(path: Path) -> dict:
    return {
        "hash": file_hash(path)
    }


def fingerprint_stat(path: Path) -> dict:
    stat = path.stat()
    return {
        "mtime": stat.st_mtime,
        "ctime": stat.st_ctime,
        "size": stat.st_size,
    }


class FileSet:
    """ A collection of fingerprinted files.

    Currently the fingerprint is only a SHA256 hash.

    Two constructors are provided for building an instance from either files that
    live on the filesystem, via :py:meth:`from_filesystem` or from a persisted
    cached record available from the :py:meth:`from_cache`.
    The filesystem version actively reads all inputs files at object creation
    time, so this can be costly, especially if repeated.
    """
    # XXX: Do we need both files (set), and file_meta (dict)?  Try to make this work with just files_meta
    __slots__ = ["files", "files_meta", "get_fingerprint"]

    def __init__(self, fingerprint=fingerprint_hash):
        self.files = set()
        self.files_meta = {}
        self.get_fingerprint = fingerprint

    def __eq__(self, other: FileSet) -> bool:
        return self.files_meta == other.files_meta

    def __ne__(self, other: FileSet) -> bool:
        return self.files_meta != other.files_meta

    '''
    def __iadd__(self, other):
        self.files.update(other.files)
        self.files_meta.update(other.files_meta)

    def __add__(self, other):
        combined = FileSet()
        for item in (self, other):
            combined.files.update(item.files)
            combined.files_meta.update(item.files_meta)
        return combined
    '''

    def __iter__(self):
        return iter(self.files)

    def __len__(self):
        return len(self.files)

    @classmethod
    def from_filesystem(cls, root: Path, files: List[str] = None) -> FileSet:
        """
        Create a new FileSet instance based on a filesystem location.
        If ``files`` is None, then the entire directory is added recursively.
        """
        instance = cls()
        root = Path(root)
        if files is None:
            # Add root recursively
            instance.add_glob(root, "**/*")
            return instance
        for file_name in files:
            if file_name.endswith("/"):
                # Recursive directory walk
                instance.add_glob(root, file_name + "**/*")
            elif "*" in file_name:
                instance.add_glob(root, file_name)
            else:
                instance.add_file(root, file_name)
        return instance

    @classmethod
    def from_cache(cls, data):
        instance = cls()
        for file_name, meta in data.items():
            file_name = PurePath(file_name)
            instance.files.add(file_name)
            instance.files_meta[file_name] = meta
        return instance

    def add_file(self, root: Path, relative_path: str):
        """ Add a simple relative path to a file to the FileSet. """
        relative_path = PurePath(relative_path)
        p = root / relative_path
        if not p.is_file():
            if p.is_dir():
                # Audience: Exception text relevant to from_filesystem() caller
                raise BuildCacheException(
                    f"Expected file '{p}' is actually a directory.  If this is "
                    "correct indicate a directory with a trailing slash: "
                    f"'{p}/'")
            raise BuildCacheException(f"Missing expected file {p}")
        fp = self.get_fingerprint(p)
        self.files.add(relative_path)
        self.files_meta[relative_path] = fp

    def add_glob(self, root: Path, pattern: str):
        """ Recursively add all files matching glob pattern. """
        for p in root.glob(pattern):
            if p.is_file():
                relative_path = p.relative_to(root)

                fp = self.get_fingerprint(p)
                self.files.add(relative_path)
                self.files_meta[relative_path] = fp

    def copy_all(self, src_dir: Path, dest_dir: Path):
        """ Copy a the given set of files from one location to another. """
        src_dir = Path(src_dir)
        dest_dir = Path(dest_dir)
        for file_name in self.files:
            src = src_dir / file_name
            dest = dest_dir / file_name
            if not dest.parent.is_dir():
                dest.parent.mkdir(parents=True)
            copy2(src, dest)


class CachedRun:
    __slots__ = ["root", "config_file", "cache_dir", "_info", "_settings", "_state"]

    STATE_NEW = "new"
    STATE_EXISTS = "exists"
    STATE_TAINT = "taint"
    STATE_DISABLED = "disabled"

    _timestamp_format = "%Y-%m-%d %H:%M:%S"

    def __init__(self, root: Path):
        self.root = root
        self.config_file = self.root / "cache.json"
        self.cache_dir = self.root / "data"
        self._settings = {}
        self._info = {}
        self._state = self.STATE_NEW

        if not self.cache_dir.is_dir():
            self.cache_dir.mkdir(parents=True)
        elif self.config_file.is_file() and self.cache_dir.is_dir():
            self._state = self.STATE_EXISTS

    def set_settings(self, cache_settings):
        self._settings.update(cache_settings)

    @property
    def cached_inputs(self):
        return FileSet.from_cache(self._info["inputs"])

    @property
    def cached_outputs(self):
        return FileSet.from_cache(self._info["outputs"])

    @property
    def exists(self):
        return self._state in (self.STATE_EXISTS,)

    @property
    def is_new(self):
        return self._state == self.STATE_NEW

    @property
    def is_expired(self):
        timeout = self._settings["timeout"]
        if timeout is None:
            return False
        try:
            create_time = self._info["timestamp"]
            timeout = timedelta(seconds=timeout)
            expired = create_time + timeout
            return datetime.now() > expired
        except KeyError:  # no cover
            raise   # XXX: For initial testing
            # If anything about the info/settings were missing; assume expired
            return True

    @property
    def is_disabled(self):
        return self._state == self.STATE_DISABLED

    def inputs_identical(self, inputs: FileSet) -> bool:
        return self.cached_inputs == inputs

    def dump(self):
        def map_keys(d):
            return {str(k): v for k, v in d.items()}
        mode = "w"
        meta = dict(self._info)
        inputs = meta.pop("inputs")
        outputs = meta.pop("outputs")
        data = {
            "settings": self._settings,
            "timestamp": datetime.now().strftime(self._timestamp_format),
            "meta": meta,
            "state": {
                "inputs": map_keys(inputs),
                "outputs": map_keys(outputs),
            },
        }
        with self.config_file.open(mode) as f:
            json.dump(data, f, indent=2)

    def load(self):
        def map_keys(d):
            return {Path(k): v for k, v in d.items()}
        with self.config_file.open() as f:
            data = json.load(f)
        info = {}
        info.update(data["meta"])
        info["timestamp"] = datetime.strptime(data["timestamp"],
                                              self._timestamp_format)
        for state_type, value in data["state"].items():
            info[state_type] = map_keys(value)
        self._settings = data["settings"]
        self._info = info

    def set_cache_info(self, type: str, data: FileSet):
        assert type in ("inputs", "outputs")
        self._info[type] = data.files_meta

    def rename(self, dest):
        if dest.is_dir():
            rmtree(dest)
        self.root.rename(dest)
        # Update root incase any other operations need to take place
        self.root = dest

    def taint(self):
        cf = self.config_file
        if cf.exists():
            cf.unlink()
        self._state = self.STATE_TAINT

    def disable(self):
        self._state = self.STATE_DISABLED



# -*- coding: utf-8 -*-
"""
"""

from __future__ import absolute_import, annotations, unicode_literals

import os
import re
import sys
from os import fspath
from pathlib import Path
from typing import Callable, Optional, Sequence

from ksconf.command import ConfFileProxy
from ksconf.compat import List, Tuple
from ksconf.conf.delta import show_text_diff
from ksconf.conf.merge import merge_conf_files
from ksconf.conf.parser import PARSECONF_MID, PARSECONF_STRICT
from ksconf.consts import SMART_CREATE, SMART_NOCHANGE, SMART_UPDATE
from ksconf.hook import plugin_manager
from ksconf.layer import (DirectLayerRoot, DotDLayerRoot, LayerContext,
                          LayerFile, LayerFilter, LayerRootBase)
from ksconf.util.compare import file_compare
from ksconf.util.file import _is_binary_file, smart_copy


class LayerCombinerException(Exception):
    pass


class LayerCombiner:
    """
    Class to recursively combine layers (directories) into a single rendered output target directory.
    This is heavily used by the ``ksconf combine`` command as well as by the ``package`` command.


    Typical class use case:

    ::
        lc = LayerCombiner()

        # Setup source, either
            (1) lc.set_source_dirs()  OR
            (2) lc.set_layer_root()

    Call hierarch:

    ::

        lc.combine()                    Entry point
            -> prepare()                Directory, layer prep
                -> prepare_target_dir() Make dir; subclass handles marker here (combine CLI)
            -> pre_combine_inventory()  Hook for pre-processing (or alerting) the set of files to combine
            -> combine_files()          Main worker function
            -> post_combine()           Optional, cleanup leftover files
    """

    # Note this is case sensitive.  Don't be lazy, name your files correctly  :-)
    conf_file_re = re.compile(r"([a-z_-]+\.conf|(default|local)\.meta)$")
    spec_file_re = re.compile(r"\.conf\.spec$")

    filetype_handlers: List[Tuple[Callable, Callable]] = []

    def __init__(self,
                 follow_symlink: bool = False,
                 banner: str = "",
                 dry_run: bool = False,
                 quiet: bool = False):
        self.layer_root: Optional[LayerRootBase] = None
        self.context = LayerContext()
        self.layer_filter = LayerFilter()
        self.banner = banner
        self.dry_run = dry_run
        self.quiet = quiet

        self.context.follow_symlink = follow_symlink

        # Internal tracking variables
        self.layer_names_all = set()
        self.layer_names_used = set()

        # Not a great long-term design, but good enough for initial conversion from command-based design
        self.stdout = sys.stdout
        self.stderr = sys.stderr

    @classmethod
    def register_handler(cls, regex_match):
        """ Decorator that registers a new file type handler.  The handler is
        used if a file name matches a regex.  Regex 'search' mode is used.
        """
        cre = re.compile(regex_match)

        def match_f(file: Path):
            if cre.search(file.name):
                return True
            return False

        def wrapper(handle_f):
            cls.filetype_handlers.append((match_f, handle_f))
            return handle_f

        return wrapper

    def log(self, message):
        # For historic reasons (idk) we send all messages to stderr, and diffs to stdout
        if message:
            self.stderr.write(f"{message}\n")

    def debug(self, message):
        if not self.quiet:
            self.log(message)

    def set_source_dirs(self, sources: List[Path]):
        self.layer_root = DirectLayerRoot(context=self.context)
        for src in sources:
            self.layer_root.add_layer(Path(src))

    def set_layer_root(self, root: LayerRootBase.Layer):
        layer_root = DotDLayerRoot(context=self.context)
        layer_root.set_root(root)
        self.layer_root = layer_root

    def add_layer_filter(self, action, pattern):
        self.layer_filter.add_rule(action, pattern)

    def combine(self, target: Path, *, hook_label=""):
        """
        Combine layers into ``target`` directory.
        Any ``hook_label`` given will be passed to the plugin system via the
        ``usage`` field.
        """
        layer_root = self.layer_root
        if layer_root is None:
            raise TypeError("Call either set_source_dirs() or set_layer_root() before calling combine()")
        self.prepare(target)
        # Build a common tree of all src files.
        src_file_listing = layer_root.list_files()
        src_file_listing = self.pre_combine_inventory(target, src_file_listing)
        self.combine_files(target, src_file_listing)
        self.post_combine(target)

        plugin_manager.hook.post_combine(target=target, usage=hook_label)

    def prepare(self, target: Path):
        """ Start the combine process.  This includes directory checking,
        applying layer filtering, and marker file handling. """
        layer_root, layer_filter = self.layer_root, self.layer_filter
        target = Path(target)

        self.prepare_target_dir(target)

        self.layer_names_all.update(layer_root.list_layer_names())
        if layer_root.apply_filter(layer_filter):
            self.layer_names_used.update(layer_root.list_layer_names())
        else:
            self.layer_names_used.update(self.layer_names_all)

    def prepare_target_dir(self, target: Path):
        """ Hook to ensure destination directory is ready for use.  This can be overridden
        to adder marker file handling for use cases that need it (e.g., the 'combine' command)
        """
        if not target.is_dir():
            target.mkdir()

    def pre_combine_inventory(self, target: Path, src_files: List[LayerFile]) -> Sequence[LayerFile]:
        """ Hook point for pre-processing before any files are copied/merged """
        del target
        return src_files

    def post_combine(self, target):
        """ Hook point for post-processing after all copy/merge operations have been completed. """
        del target

    def combine_files(self, target: Path, src_files: List[LayerFile]):
        layer_root = self.layer_root
        for src_file in sorted(src_files):
            do_copy = True
            sources = list(layer_root.get_file(src_file))
            try:
                dest_fn = sources[0].logical_path
            except IndexError:
                raise LayerCombinerException(f"File disappeared during execution?  {src_file}\n")

            dest_path: Path = target / dest_fn

            # Make missing destination folder, if missing
            dest_dir: Path = dest_path.parent
            if not dest_dir.is_dir() and not self.dry_run:
                dest_dir.mkdir(parents=True)

            # Determine handling method based on source count and filename pattern
            if len(sources) > 1:
                for matcher, handler in self.filetype_handlers:
                    if matcher(dest_fn):
                        handler(self, dest_path, sources, self.dry_run)
                        do_copy = False

            if do_copy:
                # self.debug((f"Considering {fspath(dest_fn):50}  NON-CONF Copy from source:  "
                #             f"{sources[-1].physical_path!r}")
                # Always use the last file in the list (since last directory always wins)
                src_file = sources[-1].resource_path
                if self.dry_run:
                    if dest_path.is_file():
                        if file_compare(src_file, dest_path):
                            smart_rc = SMART_NOCHANGE
                        else:
                            if (_is_binary_file(src_file) or _is_binary_file(dest_path)):
                                # Binary files.  Can't compare...
                                smart_rc = "DRY-RUN (NO-DIFF=BIN)"
                            else:
                                show_text_diff(self.stdout, dest_path, src_file)
                                smart_rc = "DRY-RUN (DIFF)"
                    else:
                        smart_rc = "DRY-RUN (NEW)"
                else:
                    smart_rc = smart_copy(src_file, dest_path)
                if smart_rc != SMART_NOCHANGE:
                    self.debug(f"Copy <{smart_rc}>   {fspath(dest_path):50}  from {src_file}")
                del src_file


# registration decorator
register_handler = LayerCombiner.register_handler


#
# Combine handlers for specific file types
#

@register_handler(r"([a-z_-]+\.conf|(default|local)\.meta)$")
def handle_merge_conf_files(combiner: LayerCombiner,
                            dest_path: Path,
                            sources: List[LayerFile],
                            dry_run):
    """
    Handle merging two or more ``.conf`` files.
    """
    sources_physical = [fspath(p.physical_path) for p in sources]
    # Note that latest mtime logic is handled by merge_conf_files()
    srcs = []
    try:
        # Handle merging conf files
        dest = ConfFileProxy(dest_path, "r+", parse_profile=PARSECONF_MID)
        srcs = [ConfFileProxy(s.resource_path, "r", parse_profile=PARSECONF_STRICT) for s in sources]
        # combiner.debug(f"Considering {dest_fn:50}  CONF MERGE from source:  "
        #                f"{1!sources[0].physical_path}")
        smart_rc = merge_conf_files(dest, srcs, dry_run=dry_run,
                                    banner_comment=combiner.banner)
        if smart_rc != SMART_NOCHANGE:
            combiner.debug(f"Merge <{smart_rc}>   {fspath(dest_path):50}  from {sources_physical!r}")
        return smart_rc
    finally:
        # Protect against any dangling open files:  (ResourceWarning: unclosed file)
        dest.close()
        for src in srcs:
            src.close()


@register_handler(r"\.conf\.spec$")
def handle_spec_concatenate(combiner: LayerCombiner,
                            dest_path: Path,
                            sources: List[LayerFile],
                            dry_run):
    """
    Concatenate multiple ``.spec`` files.  Likely a ``README.d`` situation.
    """
    sources_physical = []
    combined_content = ""
    last_mtime = max(src.mtime for src in sources)
    for src in sources:
        sources_physical.append(src.physical_path)
        content = src.resource_path.read_text()
        if not content.endswith("\n"):
            content += "\n"
        combined_content += content
    smart_rc = SMART_CREATE
    if dest_path.is_file():
        dest_content = dest_path.read_text()
        if dest_content == combined_content:
            smart_rc = SMART_NOCHANGE
        else:
            smart_rc = SMART_UPDATE

    if not dry_run:
        dest_path.write_text(combined_content)

    os.utime(dest_path, (last_mtime, last_mtime))
    if smart_rc != SMART_NOCHANGE:
        combiner.debug(f"Concatenate <{smart_rc}>   {fspath(dest_path):50}  from {sources_physical!r}")
    return smart_rc



""" ksconf.command:

Helpers functions and classes in support of the actual commands that live under
``ksconf.commands.*``.

Note that ``ksconf.commands`` is a namespace package, which can be contributed
to by multiple python packages (technically called "distributions").  Because of
this, there can be no __init__.py, which is where this content logically belongs.

"""

from __future__ import absolute_import, unicode_literals

import argparse
import logging
import os
import re
import sys
import textwrap
import time
from argparse import ArgumentParser, ArgumentTypeError
from collections import namedtuple
from io import StringIO, open
from textwrap import dedent
from typing import Optional, TextIO
from warnings import warn

from ksconf.compat import cache
from ksconf.conf.parser import (ConfParserException, ParserConfig,
                                detect_by_bom, parse_conf, smart_write_conf,
                                write_conf)
from ksconf.consts import (EXIT_CODE_BAD_ARGS, EXIT_CODE_BAD_CONF_FILE,
                           EXIT_CODE_NO_SUCH_FILE, SMART_CREATE, SmartEnum)
from ksconf.hook import plugin_manager
from ksconf.hookspec import KsconfPluginWarning
from ksconf.util import debug_traceback

__all__ = [
    "KsconfCmd",
    "ConfDirProxy",
    "ConfFileProxy",
    "ConfFileType",
    "dedent",
    "get_all_ksconf_cmds",
    "get_entrypoints",
    "add_splunkd_access_args",
    "add_splunkd_namespace",
]


class ConfDirProxy:
    def __init__(self, name, mode, parse_profile=None):
        self.name = name
        self._mode = mode
        self._parse_profile = parse_profile

    def get_file(self, relpath):
        path = os.path.join(self.name, relpath)
        return ConfFileProxy(path, self._mode, parse_profile=self._parse_profile, is_file=True)


class ConfFileProxy:
    def __init__(self, name: str, mode: str, stream: Optional[TextIO] = None,
                 *,
                 parse_profile: Optional[ParserConfig] = None,
                 is_file: Optional[bool] = None):
        self.name = os.fspath(name)
        self._mode = mode
        if is_file is not None:
            self._is_file = is_file
        elif stream:
            self._is_file = False
        elif self.writable() or os.path.isfile(name):
            self._is_file = True
        else:
            self._is_file = False
        self._stream = stream

        # Not sure if there's a good reason to keep a copy of the data locally?
        self._data = None
        self._parse_profile = parse_profile or {}

    def __del__(self):
        if self.is_file():
            self.close()

    def readable(self):
        return "r" in self._mode

    def writable(self):
        return "+" in self._mode or "w" in self._mode

    def is_file(self):
        # Is "seekable" a more appropriate distinction?  (match IOBase)
        return self._is_file

    def _type(self):    # pragma: no cover  (only used in exceptions)
        if self._is_file:
            return "file"
        else:
            return "stream"

    def close(self):
        if self._stream:
            if not self._stream.closed:
                try:
                    self._stream.close()
                finally:
                    del self._stream
        self._stream = None

    def reset(self):
        if self._data is not None:
            self._data = None
            if self.is_file():
                self.close()
            else:
                self.stream.seek(0)

    def set_parser_option(self, **kwargs):
        """ Setting a key to None will remove that setting. """
        profile = dict(self._parse_profile)
        for (k, v) in kwargs.items():
            if v is None:
                if k in profile:
                    del profile[k]
            else:
                cv = profile.get(k, None)
                if cv != v:
                    profile[k] = v
        if self._parse_profile != profile:
            self._parse_profile = profile
            self.reset()

    @property
    def stream(self):
        if self._stream is None:
            self._stream = open(self.name, self._mode)
        return self._stream

    @property
    def data(self):
        if self._data is None:
            self._data = self.load()
        return self._data

    @property
    def mtime(self):
        if self._is_file:
            return os.stat(self.name).st_mtime
        else:
            return time.time()

    def load(self, profile=None):
        if not self.readable():
            # Q: Should we mimic the exception caused by doing a read() on a write-only file object?
            raise ValueError(f"Unable to load() from {self._type()} with mode '{self._mode}'")
        parse_profile = dict(self._parse_profile)
        if profile:
            parse_profile.update(profile)
        data = parse_conf(self.stream, profile=parse_profile)
        return data

    def dump(self, data, **kwargs) -> SmartEnum:
        if not self.writable():      # pragma: no cover
            raise ValueError(f"Unable to dump() to {self._type()} with mode '{self._mode}'")
        # Feels like the right thing to do????  OR self._data = data
        self._data = None
        # write vs smart write here ----
        if self._is_file:
            self.close()
            return smart_write_conf(self.name, data, **kwargs)
        else:
            write_conf(self._stream, data, **kwargs)
            return SMART_CREATE

    def unlink(self):
        # Eventually this could trigger some kind of backup or recovery mechanism
        self.close()
        return os.unlink(self.name)

    '''
    def backup(self, bkname=None):
        # One option:  Write this file directly to the git object store.  Just need to store some
        # kind of index to allow the users to pull it back.   (Sill, would need of fall-back
        # mechanism).  Git shouldn't be a hard-dependency
        raise NotImplementedError

    def checksum(self, hash="sha256"):
        raise NotImplementedError
    '''


class ConfFileType:
    """Factory for creating conf file object types;  returns a lazy-loader ConfFile proxy class

    Started from FileType() and then changed everything.   With our use case, it's often
    necessary to delay writing, or read before writing to a conf file (depending on whether or not
    --dry-run mode is enabled, for example.)

    Instances of FileType are typically passed as type= arguments to the
    ArgumentParser add_argument() method.


    :param mode: How the file is to be opened.  Accepts "r", "w", and "r+".
    :type mode: str
    :param action: Determine how much work should be handled during argument parsing vs handed off
                   to the caller.  Supports 'none', 'open', 'load'.  Full descriptions below.
    :type action: str
    :param parse_profile: parsing configuration settings passed along to the parser
    :param bool accept_dir: Should the CLI accept a directory of config files instead of an
                            individual file.  Defaults to `False`.

    **Values for action**

    ========    =============
    Action      Description
    ========    =============
    ``none``    No preparation or testing is done on the filename.
    ``open``    Ensure the file exists and can be opened.
    ``load``    Ensure the file can be opened and parsed successfully.
    ========    =============


    Once invoked, instances of this class will return a :class:`ConfFileProxy` object, or a
    :class:`ConfDirProxy` object if a directory is passed in via the CLI.
    """

    def __init__(self, mode='r', action="open",
                 parse_profile: Optional[ParserConfig] = None,
                 accept_dir: bool = False):
        self._mode = mode
        self._action = action
        self._parse_profile = parse_profile or {}
        self._accept_dir = accept_dir

    def __call__(self, string):
        # the special argument "-" means sys.std{in,out}
        if string == '-':
            if 'r' in self._mode:
                cfp = ConfFileProxy("<stdin>", "r", stream=sys.stdin, is_file=False)
                if self._action == "load":
                    try:
                        cfp.data  # noqa: F841
                    except ConfParserException as e:
                        raise ArgumentTypeError(f"failed to parse <stdin>: {e}")
                return cfp
            elif 'w' in self._mode:
                return ConfFileProxy("<stdout>", "w", stream=sys.stdout, is_file=False)
            else:
                raise ValueError(f'argument "-" with mode {self._mode}')
        if self._accept_dir and os.path.isdir(string):
            return ConfDirProxy(string, self._mode, parse_profile=self._parse_profile)
        if self._action == "none":
            return ConfFileProxy(string, self._mode, parse_profile=self._parse_profile)
        else:
            try:
                # Another possible option is using:  seekable()  but that only works ONCE the stream is open
                if os.path.isfile(string) and not string.startswith("/dev/fd/"):
                    encoding = detect_by_bom(string)
                    stream = open(string, self._mode, encoding=encoding)
                    cfp = ConfFileProxy(string, self._mode, stream=stream,
                                        parse_profile=self._parse_profile, is_file=True)
                else:
                    # Could be an explicit link to /dev/stdin; or /dev/fd/63; a bash <(cmd) input
                    # Assume UTF-8 here because that's the default encoding expected by Splunk
                    stream = open(string, self._mode, encoding="utf-8")
                    cfp = ConfFileProxy(string, self._mode, stream=stream,
                                        parse_profile=self._parse_profile, is_file=False)
                if self._action == "load":
                    # Force file to be parsed by accessing the 'data' property
                    cfp.data  # noqa: F841
                return cfp
            except IOError as e:
                debug_traceback()
                raise ArgumentTypeError(f"can't open '{string}': {e}")
            except ConfParserException as e:
                debug_traceback()
                raise ArgumentTypeError(f"failed to parse '{string}': {e}")
            except TypeError as e:
                debug_traceback()
                raise ArgumentTypeError(f"Parser config error '{string}': {e}")

    def __repr__(self):     # pragma: no cover
        args = self._mode, self._action, self._parse_profile
        args_str = ', '.join(repr(arg) for arg in args if arg != -1)
        return f"{type(self).__name__}({args_str})"


class DescriptionFormatterNoReST(argparse.HelpFormatter):
    @staticmethod
    def strip_simple_rest(s):
        # No handling of embedded backticks for now...  let's keep this simple
        # Replace literals ``X`` with single quote version:  'X'
        s = re.sub(r'``([^`]*)``', r"'\1'", s)
        # Handle simple references and other named inline markups
        # Handle explicitly named entry first, Use "<title>"
        s = re.sub(r':[a-z][a-z-]{1,10}[a-z]:`[^`]*? <([^`>]*)>\s*`', r"\1", s)
        # Just keep the content of the ref name as-is.  (no "<title>")
        s = re.sub(r':[a-z][a-z-]{1,10}[a-z]:`([^`]*)`', r"'\1'", s)
        return s

    def _fill_text(self, text, width, indent):
        text = self.strip_simple_rest(text)
        text = self._whitespace_matcher.sub(' ', text).strip()
        return textwrap.fill(text, width, initial_indent=indent,
                             subsequent_indent=indent)

    def _split_lines(self, text, width):
        text = self.strip_simple_rest(text)
        text = self._whitespace_matcher.sub(' ', text).strip()
        return textwrap.wrap(text, width)


class DescriptionHelpFormatterPreserveLayout(DescriptionFormatterNoReST):

    def _fill_text(self, text, width, indent):
        text = self.strip_simple_rest(text)
        # Looks like this one is ONLY used for the top-level description
        return ''.join([indent + line for line in text.splitlines(True)])


class KsconfCmdReadConfException(Exception):
    def __init__(self, rc):
        self.returncode = rc


class KsconfCmd:
    """ Ksconf command specification base class. """
    help: Optional[str] = None
    description: Optional[str] = None
    format = "default"
    maturity = "alpha"
    version_extra: Optional[str] = None

    def __init__(self, name):
        self.name = name.lower()
        # XXX:  Add logging support.  Find clean lines between logging and UI/console output.
        self.logger = logging.getLogger(f"ksconf.cmd.{self.name}")
        self.stdin = sys.stdin
        self.stdout = sys.stdout
        self.stderr = sys.stderr
        self.parse_profile = {}

    @classmethod
    def _handle_imports(cls):
        """ Child classes can override this to provide a save place to import 3rd party modules.

        Any ImportErrors thrown here are handled as a special case.  Allow the user of external
        module without killing the entire suite due to one missing library. """
        pass

    '''
    def redirect_io(self, stdin=None, stdout=None, stderr=None):
        if stdin is not None:
            self.stdin = stdin
        if stdout is not None:
            self.stdout = stdout
        if stderr is not None:
            self.stderr = stderr
    '''

    def exit(self, exit_code):
        """ Allow overriding for unittesting or other high-level functionality, like an
        interactive interface. """
        sys.exit(exit_code)

    def add_parser(self, subparser):
        # Passing in the object return by 'ArgumentParser.add_subparsers()'
        kwargs = {
            "help": self.help,
            "description": self.description,
        }
        if self.format == "manual":
            kwargs["formatter_class"] = DescriptionHelpFormatterPreserveLayout
        else:
            kwargs["formatter_class"] = DescriptionFormatterNoReST
        self.parser = subparser.add_parser(self.name, **kwargs)
        self.parser.set_defaults(funct=self.launch)
        self.register_args(self.parser)
        plugin_manager.hook.ksconf_cli_modify_argparse(parser=self.parser, name=self.name)

    def register_args(self, parser: ArgumentParser):        # pragma: no cover
        """ This function in passed the """
        raise NotImplementedError

    def launch(self, args):
        """ Handle flow control between pre_run() / run() / post_run() """
        # If this fails, exception is passed up, no handling errors/logging done here.
        return_code = self.pre_run(args)
        if return_code:
            return return_code

        exc = None
        try:
            return_code = self.run(args)
        except KsconfCmdReadConfException as e:
            return_code = e.returncode
        except BrokenPipeError:    # pragma: no cover
            try:
                self.stderr.write("Broken pipe\n")
            except Exception:
                pass
        except BaseException:     # pragma: no cover
            exc = sys.exc_info()
            raise
        finally:
            # No matter what, if run() was called, so is post_run()
            self.post_run(args, exc)
        return return_code

    def pre_run(self, args):
        """ Optional pre-run hook.
        Any exceptions or non-0 return code, will prevent run()/post_run() from being called. """
        pass

    def run(self, args):    # pragma: no cover
        """ Actual works happens here.  Return code should be an EXIT_CODE_* from consts. """
        raise NotImplementedError

    def post_run(self, args, exec_info=None):
        """ Optional custom clean up method.
        Always called if run() was.  The presence of exc_info indicates failure. """
        pass

    # Helper functions

    def _parse_conf(self, path, mode, profile):
        p = dict(self.parse_profile)
        if profile:
            p.update(profile)
        if path == "-":
            cfp = ConfFileProxy("<stdin>", "r", stream=sys.stdin, is_file=False)
        else:
            encoding = detect_by_bom(path)
            stream = open(path, mode, encoding=encoding)
            cfp = ConfFileProxy(path, mode, stream=stream,
                                parse_profile=p, is_file=True)
        d = cfp.data
        del d
        return cfp

    def parse_conf(self, path: str, mode: str = "r",
                   profile: Optional[ParserConfig] = None,
                   raw_exec: bool = False) -> ConfFileProxy:
        if raw_exec:
            return self._parse_conf(path, mode, profile)
        try:
            return self._parse_conf(path, mode, profile)
        except IOError as e:
            # debug_traceback()
            self.stderr.write(f"can't open '{path}': {e}\n")
            raise KsconfCmdReadConfException(EXIT_CODE_NO_SUCH_FILE)
        except ConfParserException as e:
            # debug_traceback()
            self.stderr.write(f"Failed to parse '{path}':  {e}\n")
            raise KsconfCmdReadConfException(EXIT_CODE_BAD_CONF_FILE)
        except TypeError as e:
            # debug_traceback()
            self.stderr.write(f"Parser config error '{path}': {e}\n")
            # I guess bad conf file.... can't remember what this one is for.
            raise KsconfCmdReadConfException(EXIT_CODE_BAD_CONF_FILE)

    def parse_extra_vars(self, vars: str, arg_name="argument") -> dict:
        """ Argument can be either a string, or a @file """
        # XXX: Add more error checking and better user feedback!
        if vars.startswith("@"):
            filename = vars[1:]
            payload = ""
        else:
            filename = ""
            payload = vars

        # I'm not sure this really makes sense, but why not?
        if filename.endswith(".conf"):
            return self.parse_conf(filename)

        if not payload:
            payload = open(filename).read()

        try:
            import json
            data = json.loads(payload)
            if not isinstance(data, dict):
                self.stderr.write(f"Provided {arg_name} must contain a dict, not a {type(data)}\n")
                return self.exit(EXIT_CODE_BAD_ARGS)
        except ValueError:
            # Couldn't parse as JSON; move on ...
            pass

        try:
            import yaml
            if filename:
                with open(filename) as fp:
                    data = yaml.safe_load(fp)
            else:
                data = yaml.safe_load(StringIO(payload))

            if not isinstance(data, dict):
                self.stderr.write(f"Provided {arg_name} must contain a dict, not a {type(data)}\n")
                return self.exit(EXIT_CODE_BAD_ARGS)
            return data
        except ValueError:
            # Not YAML?....
            pass
        except ImportError:
            # PyYAML not installed
            self.stderr.write("Unable to load from YAML as PyYAML not installed")
            pass

        if filename:
            self.stderr.write(f"Unable to parse {arg_name} from file {filename}\n")
        else:
            self.stdout.write(f"Unable to parse {arg_name} from:\n{payload}\n\n")
        self.exit(EXIT_CODE_BAD_ARGS)


def add_splunkd_access_args(parser: ArgumentParser) -> ArgumentParser:
    parser.add_argument("--url", default="https://localhost:8089",
                        help="URL of Splunkd.  Default:  %(default)s")
    parser.add_argument("--user", default="admin",
                        help="Login username Splunkd.  Default:  %(default)s")
    parser.add_argument("--pass", dest="password", default="changeme",
                        help="Login password Splunkd.  Default:  %(default)s")
    parser.add_argument("-k", "--insecure", action="store_true", default=False,
                        help="Disable SSL cert validation.")
    parser.add_argument("--session-key", default=None,
                        help="Use an existing session token instead of using a "
                             "username and password to login.")
    parser.add_argument("--token", default=None,
                        help="Use a Splunk token (bearer token)")
    return parser


def add_splunkd_namespace(parser: ArgumentParser) -> ArgumentParser:
    parser.add_argument("--app", default="search",
                        help="Set the namespace (app name) for the endpoint. "
                        "Default %(default)s")
    parser.add_argument("--owner", default="nobody",
                        help="Set the user who owns the content.  "
                             "The default of 'nobody' works well for app-level sharing.")
    parser.add_argument("--sharing", default="app", choices=["user", "app", "global"],
                        help="Set the sharing mode.  Defaults to 'app'.")
    return parser


def add_file_handler(parser: ArgumentParser) -> ArgumentParser:
    from ksconf.layer import layer_file_factory
    handlers = layer_file_factory.list_available_handlers()

    parser.add_argument("--enable-handler", action="append", default=[], choices=handlers,
                        help="Enable optional file handling support")
    return parser


def _get_importlib_entrypoints(group, name=None) -> list:
    # Using a backport library to get Python 3.10 EntryPoints.select() functionality.
    # But practically, if the backport is available, use it.  It's likely newer than stdlib.
    try:
        from importlib_metadata import entry_points
    except ImportError:
        from importlib.metadata import entry_points
    return entry_points(group=group, name=name)


def _get_fallback(group, name=None):
    from ksconf.setup_entrypoints import get_entrypoints_fallback
    entrypoints = get_entrypoints_fallback(group)
    if name is None:
        return entrypoints
    else:
        return entrypoints[name]


__get_entity_resolvers = [
    _get_importlib_entrypoints,
    _get_fallback,
]

if "ksconf_cmd" in os.environ.get("KSCONF_DISABLE_PLUGINS", ""):    # pragma: no cover
    # Only use the fallback built in mechanism.  This is helpful when unittesting and building docs
    # as we don't want to accidentally document/test code from other packages.
    __get_entity_resolvers = [_get_fallback]


# This caching is *mostly* beneficial for unittest CLI testing
@cache
def get_entrypoints(group, name=None):

    for resolver in list(__get_entity_resolvers):
        results = None
        try:
            results = resolver(group, name=name)
        except ImportError:    # pragma: no cover
            __get_entity_resolvers.remove(resolver)
        if results:
            return results
        # Otherwise try next technique ...


KsconfCmdEntryPoint = namedtuple("KsconfCmdEntryPoint", ["name", "entry", "cmd_cls", "error"])


def get_all_ksconf_cmds(on_error="warn"):
    for (name, entry) in get_entrypoints("ksconf_cmd").items():
        try:
            cmd_cls = entry.load()
        except (ImportError, NameError, SyntaxError) as e:
            if on_error == "warn":
                warn(f"Unable to load entrypoint for {name}.  Disabling.\n"
                     f"Base exception {e}.", KsconfPluginWarning)
            elif on_error == "return":
                error = f"Internal error:  {e}"
                yield KsconfCmdEntryPoint(name, entry, None, error)
            else:
                raise e
            continue
        if not issubclass(cmd_cls, KsconfCmd):
            msg = "Issue loading class for entrypoint:  Disabling.\n" \
                  f"{entry!r} is not derived from KsconfCmd.  "
            if on_error == "warn":
                warn(msg, KsconfPluginWarning)
            elif on_error == "return":
                yield KsconfCmdEntryPoint(name, entry, cmd_cls, msg)
            else:
                raise RuntimeError(msg)
            continue
        try:
            cmd_cls._handle_imports()
        except ImportError as e:
            module = e.name
            if on_error == "warn":
                warn(f"Unable to load external modules for {name}.  Disabling.  "
                     f"{module}.", KsconfPluginWarning)
            elif on_error == "return":
                error = f"Missing 3rd party module:  {module}"
                yield KsconfCmdEntryPoint(name, entry, cmd_cls, error)
            else:
                raise e
            continue
        yield KsconfCmdEntryPoint(name, entry, cmd_cls, None)



# -*- coding: utf-8 -*-
""" Splunk App content inventory and signature management

"""

from __future__ import absolute_import, annotations, unicode_literals

import hashlib
import json
from dataclasses import asdict, dataclass, field
from os import fspath
from pathlib import Path, PurePosixPath
from typing import Callable, Iterable, Optional, Union

from ksconf.archive import extract_archive
from ksconf.compat import List
from ksconf.consts import _UNSET, MANIFEST_HASH, UNSET
from ksconf.util.file import atomic_open, file_hash, relwalk


class AppArchiveError(Exception):
    pass


class AppArchiveContentError(Exception):
    """ Problem with the contents of an archive """
    pass


class AppManifestStorageError(Exception):
    pass


class AppManifestStorageInvalid(AppManifestStorageError):
    pass


@dataclass(order=True)
class AppManifestFile:
    """
    Manifest entry for a single file contained within an app.

    You probably don't want this class.  Use :py:class:`AppManifest` instead.
    """
    path: PurePosixPath
    mode: int
    size: int
    hash: Optional[str] = None

    def content_match(self, other: AppManifestFile):
        return self.hash == other.hash

    def to_dict(self):
        d = asdict(self)
        d["mode"] = f"0{self.mode:03o}"
        d["path"] = fspath(self.path)
        return d

    @classmethod
    def from_dict(cls, data: dict) -> AppManifestFile:
        # v1: 'mode' is int
        # v2: 'mode' is octal string
        mode = data["mode"]
        try:
            mode = int(mode, 8)
        except TypeError:
            if not isinstance(mode, int):
                raise ValueError(f"Unable to handle mode value {mode!r}")
        return cls(PurePosixPath(data["path"]), mode, data["size"], data["hash"])


FileFilterFunction = Callable[[PurePosixPath], bool]


@dataclass
class AppManifest:
    """
    Manifest of a Splunk app.  It contains the signatures of contained files and
    optionally a hash signature of app content.

    This is quite very different than a tarball hash, which includes "noise",
    like file modification time and possibly tarball creation time.  These
    factors make comparison more expensive.  The goal of this class is the
    ability to capture an app's content "fingerprint" and quickly determine if
    another app is the same or not.  And to compare apps across equally between
    tarballs, expanded folders, or a serialized capture at a point in time.

    Build instances using:

    * :py:meth:`from_tarball` - from a Splunk ``.spl`` or ``.tar.gz`` file.
    * :py:meth:`from_filesystem` - from an extracted Splunk app directory
    * :py:meth:`from_dict` - primarily for json serialization from :py:meth:`to_dict`.
    """
    name: Optional[str] = None
    source: Union[str, Path, None] = None
    hash_algorithm: str = field(default=MANIFEST_HASH)
    _hash: Union[str, _UNSET] = field(default=UNSET, init=False)
    files: List[AppManifestFile] = field(default_factory=list)

    def __eq__(self, other: AppManifest) -> bool:
        if self.name != other.name or self.hash != other.hash:
            return False
        return sorted(self.files) == sorted(other.files)

    @property
    def hash(self):
        """ Return hash, either from deserialization or local calculation. """
        if self._hash is UNSET:
            self._hash = self._calculate_hash()
        return self._hash

    @hash.deleter
    def hash(self):
        """ Reset the hash calculation.  Do this after modifying 'files'. """
        assert self.files, "Refusing to reset hash without 'files'"
        self._hash = UNSET

    def recalculate_hash(self) -> bool:
        """ Recalculate hash and indicate if hash has changed. """
        first_hash = self._hash
        assert first_hash is not UNSET, "Hash has not been calculated or provided"
        assert first_hash is not None, "Hash cannot be calculated"
        del self.hash
        return first_hash != self.hash

    def _calculate_hash(self) -> Optional[str]:
        """ Build unique hash based on file content """
        # Path sort order notes.  Sorting based on Path objects is different
        # than textual sorting, consider:
        #   README.txt
        #   README/inputs.conf.spec
        # Sorting based on path is equivalent to sorting tuples of path components
        # Like doing sort(key=lambda s: s.path.split("/"))
        parts = []
        for f in sorted(self.files):
            # If one or more hash is None, then refuse to calculate hash
            if f.hash is None:
                return None
            parts.append(f"{f.hash} 0{f.mode:03o} {'/'.join(f.path.parts)}")
        parts.insert(0, self.name)
        payload = "\n".join(parts)
        h = hashlib.new(self.hash_algorithm)
        h.update(payload.encode("utf-8"))
        return h.hexdigest()

    def filter_files(self, filter: Callable[[AppManifestFile], bool]):
        """ Apply a filter function to :py:attr:`files` safely.

        Note that unlike the `filter_file` argument on :py:meth:`from_filesystem` and
        :py:meth:`from_tarball`, the :py:obj:`filter` function is given an entire
        :py:class:`AppManifestFile` object not just the file path.
        """
        if self._hash is UNSET:
            self.files = [f for f in self.files if filter(f)]
        else:
            raise TypeError("Inappropriate use of filter_files().  "
                            "This must be called before hash is calculated.")

    def drop_ds_autogen(self):
        """
        Remove place-holder files created by the deployment server from the manifest
        for the purpose of consistent hash creation.

        These files always live in ``local/app.conf`` and contain the literal
        text ``# Autogenerated file``.  Any other forms of this file are preserved.
        """
        local_app = PurePosixPath('local/app.conf')
        unwanted_hashs = {
            "a0c13b7008d9ef75d56be47cdb5ea3157f087bb7e773bd3d426a1998049e89b3"  # Nix - sha256
        }
        self.filter_files(lambda f: f.path != local_app or f.hash not in unwanted_hashs)

    @classmethod
    def from_dict(cls, data: dict) -> AppManifest:
        files = [AppManifestFile.from_dict(f) for f in data["files"]]
        o = cls(data["name"], source=data["source"],
                hash_algorithm=data["hash_algorithm"], files=files)
        o._hash = data["hash"]
        return o

    def to_dict(self):
        d = {
            "name": self.name,
            "source": fspath(self.source),
            "hash_algorithm": self.hash_algorithm,
            "hash": self.hash,
            "files": [f.to_dict() for f in self.files]
        }
        return d

    @classmethod
    def from_archive(cls, archive: Path,
                     calculate_hash=True,
                     *,
                     filter_file: Optional[FileFilterFunction] = None) -> AppManifest:
        """
        Create as new AppManifest from a tarball.  Set ``calculate_hash`` as
        False when only a file listing is needed.
        """
        manifest = cls(source=fspath(archive))
        app_names = set()
        archive = Path(archive)

        if calculate_hash:
            h_ = hashlib.new(cls.hash_algorithm)

            def gethash(content):
                h = h_.copy()
                h.update(content)
                return h.hexdigest()
        else:
            def gethash(_):
                return None

        for gaf in extract_archive(archive, lambda _: calculate_hash):
            app, relpath = gaf.path.split("/", 1)
            app_names.add(app)
            hash = gethash(gaf.payload)
            relpath = PurePosixPath(relpath)
            if filter_file is None or filter_file(relpath):
                f = AppManifestFile(relpath, gaf.mode, gaf.size, hash)
                manifest.files.append(f)
        if len(app_names) > 1:
            raise AppArchiveContentError("Found multiple top-level app names!  "
                                         f"Archive {archive} contains apps {', '.join(app_names)}")
        manifest.name = app_names.pop()
        return manifest

    @classmethod
    def from_filesystem(cls, path: Path,
                        name: Optional[str] = None,
                        follow_symlinks=False,
                        calculate_hash=True,
                        *,
                        filter_file: Optional[FileFilterFunction] = None) -> AppManifest:
        """
        Create as new AppManifest from an existing directory structure.
        Set ``calculate_hash`` as False when only a file listing is needed.
        """
        path = Path(path)
        if name is None:
            name = path.name
        manifest = cls(name, source=path)
        h_ = hashlib.new(cls.hash_algorithm)

        for (root, _, files) in relwalk(path, followlinks=follow_symlinks):
            root_path = PurePosixPath(root)
            for file_name in files:
                rel_path: PurePosixPath = root_path.joinpath(file_name)
                if filter_file is not None and not filter_file(rel_path):
                    continue
                full_path: Path = path.joinpath(root, file_name)
                st = full_path.stat()
                amf = AppManifestFile(rel_path, st.st_mode & 0o777, st.st_size)
                if calculate_hash:
                    h = h_.copy()
                    h.update(full_path.read_bytes())
                    amf.hash = h.hexdigest()
                    del h
                manifest.files.append(amf)
        return manifest

    def find_local(self) -> Iterable[AppManifestFile]:
        for f in self.files:
            if f.path.parts[0] == "local" or f.path.name == "local.meta":
                yield f

    def check_paths(self):
        """ Check for dangerous paths in the archive. """
        for file in self.files:
            path = file.path
            if path.is_absolute():
                raise AppArchiveContentError(f"Found an absolute path {file.path}")
            if ".." in path.parts or path.parts[0].startswith("~"):
                raise AppArchiveContentError(f"Found questionable path manipulation in '{file.path}'")


@dataclass
class StoredArchiveManifest:
    """
    Stored manifest for a tarball.  Typically the manifest file lives in the
    same directory as the archive.  Details around the naming, storage, and
    clean up of these persistent manifest files are managed by the caller.
    """
    archive: Path
    size: int
    mtime: float
    hash: str
    _manifest_dict: dict = field(init=False)
    _manifest: AppManifest = field(init=False, default=None)

    @property
    def manifest(self) -> AppManifest:
        # Lazy load the full manifest details later (after size/mtime/hash have been confirmed)
        if self._manifest is None:
            if self._manifest_dict:
                try:
                    self._manifest = AppManifest.from_dict(self._manifest_dict)
                except KeyError as e:
                    raise AppManifestStorageError(f"Error loading manifest {e}")
            if self._manifest.recalculate_hash():
                raise AppManifestStorageError("Manifest failed internal hash consistency test")
        return self._manifest

    @classmethod
    def from_dict(cls, data: dict) -> StoredArchiveManifest:
        o = cls(Path(data["archive"]), data["size"], data["mtime"], data["hash"])
        o._manifest_dict = data["manifest"]
        return o

    def to_dict(self):
        return {
            "archive": fspath(self.archive),
            "size": self.size,
            "mtime": self.mtime,
            "hash": self.hash,
            "manifest": self.manifest.to_dict(),
        }

    def write_json_manifest(self, manifest_file: Path):
        data = self.to_dict()
        with atomic_open(manifest_file, ".tmp", "w") as fp:
            json.dump(data, fp, indent=1)

    @classmethod
    def read_json_manifest(cls, manifest_file: Path) -> StoredArchiveManifest:
        with open(manifest_file) as fp:
            data = json.load(fp)
        return cls.from_dict(data)

    @classmethod
    def from_file(cls,
                  archive: Path,
                  manifest: AppManifest
                  ) -> StoredArchiveManifest:
        """
        Construct instance from a tarball.
        """
        stat = archive.stat()

        hash = file_hash(archive, MANIFEST_HASH)
        o = cls(archive, stat.st_size, stat.st_mtime, hash)
        o._manifest = manifest
        return o

    @classmethod
    def from_json_manifest(cls,
                           archive: Path,
                           stored_file: Path,
                           *,
                           permanent_archive: Optional[Path] = None) -> StoredArchiveManifest:
        """
        Attempt to load an archive stored manifest from ``archive`` and ``stored_file`` paths.
        If the archive has changed since the manifest was stored, then an
        exception will be raised indicating the reason for invalidation.
        """
        # XXX: Optimization: tests if archive is newer than stored_file.  No need to open/parse JSON.
        if not stored_file.exists():
            raise AppManifestStorageInvalid("No stored manifest found")

        if permanent_archive is None:
            permanent_archive = archive

        try:
            stored = cls.read_json_manifest(stored_file)

            if stored.archive != permanent_archive:
                raise AppManifestStorageInvalid(f"Archive name differs: {stored.archive!r} != {permanent_archive!r}")
            stat = archive.stat()
            if stored.size != stat.st_size:
                raise AppManifestStorageInvalid(f"Archive file size differs:  {stored.size} != {stat.st_size}")
            if abs(stored.mtime - stat.st_mtime) > 0.1:
                raise AppManifestStorageInvalid(f"Archive file mtime differs: {stored.mtime} vs {stat.st_mtime}")
        except (ValueError, KeyError) as e:
            raise AppManifestStorageError(f"Unable to load stored manifest due to {e}")
        return stored


def get_stored_manifest_name(archive: Path) -> Path:
    """
    Calculate the name of the stored manifest file based on ``archive``.
    """
    c = archive.with_name(f".{archive.name}.manifest")
    return c


def create_manifest_from_archive(
        archive_file: Path,
        manifest_file: Path,
        manifest: AppManifest) -> StoredArchiveManifest:
    """
    Create a new stored manifest file based on a given archive.
    """
    if manifest_file is None:
        manifest_file = get_stored_manifest_name(archive_file)
    sam = StoredArchiveManifest.from_file(archive_file, manifest)
    sam.write_json_manifest(manifest_file)
    return sam


def load_manifest_for_archive(
        archive: Path,
        manifest_file: Optional[Path] = None,
        *,
        read_manifest=True,
        write_manifest=True,
        permanent_archive: Optional[Path] = None,
        log_callback=print) -> AppManifest:
    """
    Load manifest for ``archive`` and create a stored copy of the manifest in
    ``manifest_file``.  On subsequent calls the manifest data stored to disk
    will be reused assuming ``manifest_file`` is up-to-date.

    File modification time and size are used to determine if ``archive`` has
    been changed since the ``manifest_file`` was written.

    If no ``manifest_file`` is provided, the default manifest naming convention
    will be applied where the ``manifest_file`` is stored in the same directory
    as ``archive``.

    If ``permanent_archive`` is provided, then we assume it is the persistent
    name and ``archive`` is a temporary resource.  In this mode, the default
    ``manifest_file`` is also based on ``permanent_archive`` not ``archive``.
    """
    archive = Path(archive)
    if permanent_archive:
        permanent_archive = Path(permanent_archive)
    manifest = None

    if manifest_file is None and (read_manifest or write_manifest):
        if permanent_archive is None:
            manifest_file = get_stored_manifest_name(archive)
        else:
            manifest_file = get_stored_manifest_name(permanent_archive)
    else:
        manifest_file = Path(manifest_file)

    if read_manifest and manifest_file.exists():
        try:
            sam = StoredArchiveManifest.from_json_manifest(archive, manifest_file,
                                                           permanent_archive=permanent_archive)
            manifest = sam.manifest
        except AppManifestStorageError as e:
            log_callback(f"Loading stored manifest failed:  {e}")

    if manifest is None:
        # log_callback(f"Calculating manifest for {archive} to be written to {manifest_file}")
        manifest = AppManifest.from_archive(archive)
        if permanent_archive:
            manifest.source = permanent_archive

        # Assume stored manifest have already undergone path checks, so existing
        # manifest are not rechecked.  As path checking is done before extraction,
        # where bad paths can actually cause damage, the check is is just a
        # precaution to promote early detection of malicious behavior.
        manifest.check_paths()

        if write_manifest:
            create_manifest_from_archive(archive, manifest_file, manifest)

    return manifest



"""

Parse and write Splunk's .conf files


According to this doc:

https://docs.splunk.com/Documentation/Splunk/7.2.3/Admin/Howtoeditaconfigurationfile

 1.  Comments must start at the beginning of a line (#)
 2.  Comments may not be after a stanza name or on an attribute's value
 3.  Supporting encoding is UTF-8 (and therefore ASCII too)

"""

import codecs
import os
import re
from enum import Enum
from io import StringIO, open
from pathlib import Path
from typing import Dict, Iterable, Iterator, List, Optional, Sequence, TextIO, Tuple, Union

from ..consts import SmartEnum
from ..types import PathType
from ..util.compare import fileobj_compare
from ..util.file import atomic_open, atomic_writer

default_encoding = "utf-8"


# Type definitions

StanzaType = Dict[str, str]
ConfType = Dict[str, StanzaType]
_StreamInput = Union[TextIO, Iterable[str]]
_StreamOutput = Union[PathType, TextIO]
_StreamNameFile = Union[PathType, _StreamInput]
ParserConfig = Dict


class Token:
    """ Immutable token object.  deepcopy returns the same object """

    def __deepcopy__(self, memo):
        memo[id(self)] = self
        return self

    # Always sort to the top of the list (should only ever be GLOBAL stanza)
    def __lt__(self, other):
        return isinstance(other, str)

    def __gt__(self, other):
        return not isinstance(other, str)


class DuplicateEnum(Enum):
    OVERWRITE = "overwrite"
    MERGE = "merge"
    EXCEPTION = "exception"


# Legacy names
DUP_OVERWRITE = DuplicateEnum.OVERWRITE
DUP_MERGE = DuplicateEnum.MERGE
DUP_EXCEPTION = DuplicateEnum.EXCEPTION

GLOBAL_STANZA = Token()

# Parsing configuration profiles

PARSECONF_MID = dict(
    keep_comments=True,
    dup_stanza=DUP_EXCEPTION,
    dup_key=DUP_OVERWRITE,
    strict=True)

PARSECONF_MID_NC = dict(
    keep_comments=False,  # No comments
    dup_stanza=DUP_EXCEPTION,
    dup_key=DUP_OVERWRITE,
    strict=True)

PARSECONF_LOOSE = dict(
    keep_comments=False,
    dup_stanza=DUP_MERGE,
    dup_key=DUP_MERGE,
    strict=False)

PARSECONF_STRICT = dict(
    keep_comments=True,
    dup_stanza=DUP_EXCEPTION,
    dup_key=DUP_EXCEPTION,
    strict=True)

PARSECONF_STRICT_NC = dict(
    keep_comments=False,  # No comment
    dup_stanza=DUP_EXCEPTION,
    dup_key=DUP_EXCEPTION,
    strict=True)


class ConfParserException(Exception):
    pass


class DuplicateKeyException(ConfParserException):
    pass


class DuplicateStanzaException(ConfParserException):
    pass


####################################################################################################
# Core parsing / conf file writing logic


def section_reader(stream: Sequence[str],
                   section_re: re.Pattern = re.compile(r'^[\s\t]*\[(.*)\]\s*$')
                   ) -> Iterator[Tuple[str, List[str]]]:
    """
    This generator break a configuration file stream into sections.  Each section contains a name
    and a list of text lines held within that section.

    Sections that have no entries must be preserved.  Any lines before the first section are send back
    with the section name of None.

    :param stream: configuration file input stream
    :type stream: file
    :param section_re: regular expression for detecting stanza headers
    :return: sections in the form of `(section_name, lines_of_text)`
    :rtype: tuple
    """
    buf: List[str] = []
    section = None
    for line in stream:
        line: str = line.rstrip("\r\n")
        match = section_re.match(line)
        if match:
            yield section, buf
            section = match.group(1)
            buf = []
        else:
            buf.append(line)
    if section or buf:
        yield section, buf


def _detect_lite(byte_str: bytes) -> Dict[str, str]:
    """ A super simple drop-in replacement for chardet.detect(byte_str) that ONLY looks for BOM or
    assumes "utf-8".
    If someday the full chardet features are needed, we could use this for optional (opportunistic)
    chardet support with this as the local fall-back function. """
    # https://stackoverflow.com/a/24370596/315892
    # UTF-8 BOM is 3 bytes, UTF-16 is 2 bytes, UTF-32 is 4 bytes
    for (enc, boms) in (
            ('utf-8-sig', (codecs.BOM_UTF8,)),
            ('utf-16', (codecs.BOM_UTF16_LE, codecs.BOM_UTF16_BE)),
            ('utf-32', (codecs.BOM_UTF32_LE, codecs.BOM_UTF32_BE))):
        if any(byte_str.startswith(bom) for bom in boms):
            return {"encoding": enc}
    return {"encoding": default_encoding}


""" # Not ready for this approach yet!  (more testing scenarios required)
try:
    from chardet import detect
else:
    _detect_lite = detect
"""


def detect_by_bom(path: PathType) -> str:
    # Refuse to consume a possibly read-once stream.  Could be a shell:  <(cmd)
    # Using an assert because this really *should* be handled by the caller, but just in case...
    assert not os.fspath(path).startswith("/dev/fd/"), "Don't bom my pipe!"

    with open(path, 'rb') as f:
        raw = f.read(4)    # will read less if the file is smaller
    encoding = _detect_lite(raw)
    return encoding["encoding"]


def cont_handler(iterable: Iterable[str],
                 continue_re: re.Pattern = re.compile(r"^(.*)\\$"),
                 breaker: str = "\n"
                 ) -> Iterator[str]:
    r"""
    Look for trailing backslashes ("`\\`") which indicate a value for an attribute is split across
    multiple lines.  This function will group such lines together, and pass all other lines through
    as-is.  Note that the continuation character must be the very last character on the line,
    trailing whitespace is not allowed.

    :param iterable: lines from a configuration file
    :type iterable: iter
    :param continue_re: regular expression to detect the continuation character
    :param breaker: joining string when combining continued lines into a single string.
           Default '`\\n`'
    :return: lines of text
    :rtype: str
    """
    buf = ""
    for line in iterable:
        mo = continue_re.match(line)
        if mo:
            buf += mo.group(1) + breaker
        elif buf:
            yield buf + line
            buf = ""
        else:
            yield line
    if buf:  # pragma: no cover
        # Weird this generally shouldn't happen.
        yield buf


def splitup_kvpairs(lines: Iterable[str],
                    comments_re: re.Pattern = re.compile(r"^\s*[#;]"),
                    keep_comments: bool = False,
                    strict: bool = False
                    ) -> Iterator[Tuple[str, str]]:
    """
    Break up 'attribute=value' entries in a configuration file.

    :param lines: the body of a stanza containing associated attributes and values
    :type lines: iter
    :param comments_re: Regular expression used to detect comments.
    :param keep_comments: Should comments be preserved in the output.  Defaults to `False`.
    :type keep_comments: bool, optional
    :param strict: Should unknown content in the stanza stop processing.  Defaults to `False`
                   allowing "junk" to be silently ignored for a best-effort parse.
    :type strict: bool, optional
    :return: iterable of (attribute,value) tuples
    """
    comment = 0
    for entry in lines:
        if comments_re.search(entry):
            if keep_comments:
                comment += 1
                yield f"#-{comment:06d}", entry
        elif "=" in entry:
            k, v = entry.split("=", 1)
            yield k.rstrip(), v.lstrip()
        elif re.search(r'^\s*\[|\]\s*$', entry):
            # ToDo:  There should be a 'loose' mode that allows this to be ignored...
            raise ConfParserException(f"Dangling stanza header:  {entry}")
        elif strict and entry.strip():
            #  if entry == "\ufeff": continue # UTF-8 BOM read as UTF-8;  But this ONLY works for PY3
            raise ConfParserException(f"Unexpected entry:  {entry!r}")


def parse_conf(stream: _StreamNameFile,
               profile: ParserConfig = PARSECONF_MID,
               encoding: Optional[str] = None
               ) -> ConfType:
    """
    Parse a .conf file.  This is a wrapper around :func:`parse_conf_stream` that allows filenames
    or stream to be passed in.

    :param stream: the path to a configuration file or open file-like object to be parsed
    :type stream: str, file
    :param profile: parsing configuration settings
    :param encoding: Defaults to the system default, (Often "utf-8")
    :return: a mapping of the stanza and attributes.
             The resulting output is accessible as [stanza][attribute] -> value
    :rtype: dict
    """
    try:
        # Placeholder stub for an eventual migration to proper class-oriented parser
        if hasattr(stream, "read"):
            return parse_conf_stream(stream, **profile)
        else:
            if not encoding:
                encoding = detect_by_bom(stream)
            # Assume it's a filename
            with open(stream, "r", encoding=encoding) as stream:
                return parse_conf_stream(stream, **profile)
    except UnicodeDecodeError as e:
        raise ConfParserException(f"Encoding error encountered: {e}")


def parse_string(s: str,
                 name: Optional[str] = None,
                 profile: ParserConfig = PARSECONF_MID) -> ConfType:
    """
    Parse a .conf file that's already in memory, as a string.
    """
    stream = StringIO(s)
    if name:
        stream.name = name
    return parse_conf_stream(stream, **profile)


def parse_conf_stream(stream: _StreamInput,
                      keys_lower: bool = False,
                      handle_conts: bool = True,
                      keep_comments: bool = False,
                      dup_stanza: DuplicateEnum = DUP_EXCEPTION,
                      dup_key: DuplicateEnum = DUP_OVERWRITE,
                      strict: bool = False) -> ConfType:
    """
    Low-level conf parsing functionality.

    Most often, either :func:`parse_conf` or :func:`parse_string` are better options.
    """
    if hasattr(stream, "name"):
        stream_name = stream.name
    else:
        stream_name = repr(stream)

    sections = {}
    # Q: What's the value of allowing line continuations to be disabled?
    if handle_conts:
        stream = cont_handler(stream)
    for section, entry in section_reader(stream):
        if section is None:
            section = GLOBAL_STANZA
        if section in sections:
            if dup_stanza == DUP_OVERWRITE:
                s = sections[section] = {}
            elif dup_stanza == DUP_EXCEPTION:
                raise DuplicateStanzaException(
                    f"Stanza [{_format_stanza(section)}] found more than once "
                    f"in config file {stream_name}")
            elif dup_stanza == DUP_MERGE:
                s = sections[section]
            else:
                raise TypeError(f"Unknown value '{dup_stanza}' for dup_stanza")
        else:
            s = sections[section] = {}
        local_stanza = {}
        for key, value in splitup_kvpairs(entry, keep_comments=keep_comments, strict=strict):
            if keys_lower:
                key = key.lower()
            if key in local_stanza:
                if dup_key in (DUP_OVERWRITE, DUP_MERGE):
                    s[key] = value
                    local_stanza[key] = value
                elif dup_key == DUP_EXCEPTION:
                    raise DuplicateKeyException(
                        f"Stanza [{_format_stanza(section)}] has duplicate "
                        f"key '{key}' in file {stream_name}")
            else:
                local_stanza[key] = value
                s[key] = value
        del s
    # If the global entry is just a blank line, drop it
    if GLOBAL_STANZA in sections:
        g = sections[GLOBAL_STANZA]
        if not g:
            # if len(g) == 1 and not g[0]:
            del sections[GLOBAL_STANZA]
    return sections


def write_conf(stream: _StreamOutput,
               conf: ConfType,
               stanza_delim: str = "\n",
               sort: bool = True,
               temp_suffix: str = ".tmp",
               mtime: Optional[float] = None):
    if not hasattr(stream, "write"):
        # Assume it's a filename
        with atomic_open(stream, temp_suffix, mode="w",
                         encoding=default_encoding) as stream:
            write_conf_stream(stream, conf, stanza_delim, sort)
            if mtime:
                os.utime(stream, (mtime, mtime))
    else:
        write_conf_stream(stream, conf, stanza_delim, sort)


def write_conf_stream(stream: TextIO,
                      conf: ConfType,
                      stanza_delim: str = "\n",
                      sort: bool = True):
    if sort:
        sorter = sorted
    else:
        sorter = list

    def write_stanza_body(stanza: dict):
        for (key, value) in sorter(stanza.items()):
            if value is None:
                value = ""
            else:
                value = str(value)
            if key.startswith("#"):
                stream.write(f"{value}\n")
            elif value:
                value = value.replace("\n", "\\\n")
                stream.write(f"{key} = {value}\n")
            else:
                # Avoid a trailing whitespace to keep the git gods happy
                stream.write(f"{key} =\n")

    keys = sorter(conf)
    while keys:
        section = keys.pop(0)
        cfg = conf[section]
        if section is not GLOBAL_STANZA:
            stream.write(f"[{section}]\n")
        write_stanza_body(cfg)
        if keys:
            stream.write(stanza_delim)


def smart_write_conf(filename: PathType,
                     conf: ConfType,
                     stanza_delim: str = "\n",
                     sort: bool = True,
                     temp_suffix: str = ".tmp",
                     mtime: Optional[float] = None) -> SmartEnum:
    """ Write conf data to a specific file, but only when necessary.
    This function is essentially the same as :func:`write_conf`, except that it
    avoids updating the file if it already exists and has the desired content.
    """
    filename = Path(filename)
    if filename.is_file():
        temp = StringIO()
        write_conf_stream(temp, conf, stanza_delim, sort)
        with open(filename, encoding=default_encoding) as dest:
            file_diff = fileobj_compare(temp, dest)
        if file_diff:
            return SmartEnum.NOCHANGE
        else:
            with atomic_writer(filename, temp_suffix) as tempfile:
                tempfile.write_text(temp.getvalue(), encoding=default_encoding)
                if mtime:
                    os.utime(tempfile, (mtime, mtime))
            return SmartEnum.UPDATE
    else:
        with atomic_open(filename, temp_suffix, mode="w") as dest:
            write_conf_stream(dest, conf, stanza_delim, sort)
        if mtime:
            os.utime(filename, (mtime, mtime))
        return SmartEnum.CREATE


def _format_stanza(stanza: Union[str, Token]) -> str:
    """ Return a more human readable stanza name."""
    if stanza is GLOBAL_STANZA:
        return "GLOBAL"
    else:
        return stanza


def _extract_comments(section: StanzaType) -> List[str]:
    """ Return a sequential list of comments REMOVED from a section dict """
    comments = []
    for key, value in sorted(section.items()):
        if key.startswith("#-"):
            comments.append(value)
            del section[key]
    return comments


def inject_section_comments(section: StanzaType,
                            prepend: Optional[str] = None,
                            append: Optional[str] = None):
    """
    Extract existing comments from section dict (in order; and remove them)
    Add in any prepend/append comments (if that comment isn't already present)
    Re-inject comments back into the section dict with fresh numbering
    """
    # Yes, this is really hacky, but the only way to make the diffs work correctly ;-(
    comments = _extract_comments(section)
    new_comments = []
    if prepend:
        for c in prepend:
            if c not in comments:
                new_comments.append(c)
    new_comments.extend(comments)
    if append:
        for c in append:
            if c not in comments:
                new_comments.append(c)
    for (i, comment) in enumerate(new_comments, 1):
        section[f"#-{i:06d}"] = comment


def _drop_stanza_comments(stanza: StanzaType) -> StanzaType:
    n = {}
    for (key, value) in stanza.items():
        if key.startswith("#"):
            continue
        n[key] = value
    return n


def conf_attr_boolean(value: Union[str, bool, int]) -> bool:
    if isinstance(value, bool):
        return value
    elif isinstance(value, str):
        value = value.lower()
        if value in ("1", "t", "y", "true", "yes"):
            return True
        elif value in ("0", "f", "n", "false", "no"):
            return False
        else:
            raise ValueError(f"Can't convert {value!r} to a boolean.")
    elif isinstance(value, int):
        if value == 0:
            return False
        elif value == 1:
            # Technically any non-0 is true; but that's unusual in typical config files.
            # Let's keep the logic the same as how stings are handled:  Only '1' as true.
            return True
        else:
            raise ValueError(f"Can't convert {value!r} to a boolean.")
    else:
        raise ValueError(f"Can't convert type {type(value)} to a boolean.")


class update_conf:
    """
    Context manager that allows for simple in-place updates to conf files.
    This provides a simple dict-like interface for easy updates.

    Usage example:

    ..  code-block:: py

            with update_conf("app.conf") as conf:
                conf["launcher"]["version"] = "1.0.2"
                conf["install"]["build"] = 33

    :param str conf_path: Path to ``.conf`` file to be edited.
    :param dict profile:  Parsing settings and strictness profile.
    :param str encoding:  encoding to use for file operations.
    :param bool make_missing:  When true, a new blank configuration file will be created
                               if ``conf_path`` is missing, otherwise an exception will be raised.
    """

    def __init__(self, conf_path: PathType,
                 profile: ParserConfig = PARSECONF_MID,
                 encoding: Optional[str] = None,
                 make_missing: bool = False):
        self.path = conf_path
        self.profile = profile
        self.encoding = encoding
        self.make_missing = make_missing
        self._data: Optional[ConfType] = None

    def __enter__(self):
        if not os.path.isfile(self.path) and self.make_missing:
            self._data = {}
        else:
            self._data = parse_conf(self.path, self.profile, self.encoding)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            # No update since an exception was raised
            return
        if self.make_missing:
            parent = os.path.dirname(self.path)
            if parent and not os.path.isdir(parent):
                os.makedirs(parent)
        if self._data is not None:
            smart_write_conf(self.path, self._data, sort=True)

    def __getitem__(self, item: str) -> StanzaType:
        return self._data[item]

    def __setitem__(self, key: str, value: StanzaType):
        self._data[key] = value
        return value

    def __contains__(self, item):
        return item in self._data

    def __iter__(self) -> Iterable[str]:
        return iter(self._data)

    def keys(self) -> List[str]:
        return list(self._data)

    def update(self, *args, **kwargs):
        self._data.update(*args, **kwargs)

    def cancel(self):
        """ Indicate that no updates were made and all processing is complete.
        An error will occur if additional read/writes are attempted. """
        self._data = None



""" SUBCOMMAND:  ``ksconf diff <CONF> <CONF>``

Usage example:

.. code-block:: sh

    ksconf diff default/props.conf default/props.conf

"""
from __future__ import absolute_import, unicode_literals

import argparse

from ksconf.command import ConfFileType, KsconfCmd, dedent
from ksconf.conf.delta import compare_cfgs, show_diff, write_diff_as_json
from ksconf.conf.parser import PARSECONF_MID_NC
from ksconf.consts import EXIT_CODE_DIFF_EQUAL, EXIT_CODE_DIFF_NO_COMMON
from ksconf.util.completers import conf_files_completer


class DiffCmd(KsconfCmd):
    help = "Compare settings differences between two .conf files ignoring spacing and sort order"
    description = dedent("""\
    Compares the content differences of two .conf files

    This command ignores textual differences (like order, spacing, and comments) and
    focuses strictly on comparing stanzas, keys, and values.  Note that spaces within
    any given value, will be compared. Multi-line fields are compared in a more traditional
    'diff' output so that long saved searches and macros can be compared more easily.
    """)
    format = "manual"
    maturity = "stable"

    def register_args(self, parser):
        parser.add_argument("conf1", metavar="CONF1", help="Left side of the comparison",
                            type=ConfFileType("r", "load", parse_profile=PARSECONF_MID_NC)
                            ).completer = conf_files_completer
        parser.add_argument("conf2", metavar="CONF2", help="Right side of the comparison",
                            type=ConfFileType("r", "load", parse_profile=PARSECONF_MID_NC)
                            ).completer = conf_files_completer
        parser.add_argument("-o", "--output", metavar="FILE",
                            type=argparse.FileType('w'), default=self.stdout,
                            help="File where difference is stored.  Defaults to standard out.")
        parser.add_argument("--detail", "-d",
                            choices=["global", "stanza", "key"], default="global",
                            help="Control the highest level for which 'replace' events may occur.")
        parser.add_argument("--comments", "-C",
                            action="store_true", default=False,
                            help="Enable comparison of comments.  (Unlikely to work consistently)")
        parser.add_argument("--format", "-f",
                            choices=["diff", "json"], default="diff",
                            help="Output file format to produce.  'diff' the the classic format used by default. "
                            "'json' is helpful when trying to review changes programmatically.")

    def run(self, args):
        ''' Compare two configuration files. '''
        args.conf1.set_parser_option(keep_comments=args.comments)
        args.conf2.set_parser_option(keep_comments=args.comments)

        cfg1 = args.conf1.data
        cfg2 = args.conf2.data

        diffs = compare_cfgs(cfg1, cfg2, replace_level=args.detail)

        if args.format == "diff":
            rc = show_diff(args.output, diffs, headers=(args.conf1.name, args.conf2.name))
            if rc == EXIT_CODE_DIFF_EQUAL:
                self.stderr.write("Files are the same.\n")
            elif rc == EXIT_CODE_DIFF_NO_COMMON:
                self.stderr.write("No common stanzas between files.\n")
            return rc
        elif args.format == "json":
            # XXX: Refactor show_diff() to separate the do-we-have-a-change logic
            write_diff_as_json(diffs, args.output, indent=4)



# Version file autogenerated by the build process
version = '0.13.5'
build = 1275
vcs_info = 'Git SHA1 acaf77d3 committed on 2023-10-20'
package_name = 'ksconf'

SHELL_VER_VARS = 'KSCONF_VERSION="0.13.5"\nKSCONF_BUILD="1275"\nKSCONF_VCS_INFO="Git SHA1 acaf77d3 committed on 2023-10-20"'

if __name__ == '__main__':
    print("# Please use 'python -m ksconf.version' instead.  This will eventually go away.")
    print(SHELL_VER_VARS)



""" SUBCOMMAND:  ``ksconf filter <CONF>``

Usage example:

.. code-block:: sh

    ksconf filter default/savedsearches.conf --stanza "My Special Search" -o my-special-search.conf

Future things to support:

 * SED-like rewriting for stanza name or key values.
 * Mini eval/query language for simple data manipulations supporting mixed used of matching modes
   on a case-by-base basis, custom logic (AND,OR,arbitrary groups), projections, and content
   rewriting.  (Should leverage custom 'combine' mini-language where possible.)

"""

from __future__ import absolute_import, unicode_literals

import argparse
import sys
from argparse import ArgumentParser
from typing import List, Tuple

from ksconf.command import ConfFileType, KsconfCmd, dedent
from ksconf.conf.parser import (PARSECONF_MID_NC, ConfParserException,
                                conf_attr_boolean, write_conf_stream)
from ksconf.consts import EXIT_CODE_BAD_CONF_FILE, EXIT_CODE_SUCCESS
from ksconf.filter import FilteredList, FilteredListWildcard, create_filtered_list
from ksconf.util.completers import conf_files_completer


def is_disabled(attributes):
    # type: (dict) -> bool
    return conf_attr_boolean(attributes.get("disabled", False))


class FilterCmd(KsconfCmd):
    help = "A stanza-aware GREP tool for conf files"
    description = dedent("""
    Filter the contents of a conf file in various ways.  Stanzas can be included
    or excluded based on a provided filter or based on the presence or value of a key.

    Where possible, this command supports GREP-like arguments to bring a familiar feel.
    """)

    # format = "manual"
    maturity = "alpha"

    def __init__(self, *args, **kwargs):
        super(FilterCmd, self).__init__(*args, **kwargs)
        self.stanza_filters = None
        self.attr_presence_filters = None

    def register_args(self, parser: ArgumentParser):
        parser.add_argument("conf", metavar="CONF", help="Input conf file", nargs="+",
                            type=ConfFileType("r", parse_profile=PARSECONF_MID_NC)
                            ).completer = conf_files_completer
        parser.add_argument("-o", "--output", metavar="FILE",
                            type=argparse.FileType('w'), default=self.stdout,
                            help="File where the filtered results are written.  "
                                 "Defaults to standard out.")
        parser.add_argument("--comments", "-C",
                            action="store_true", default=False,
                            help="Preserve comments.  Comments are discarded by default.")
        parser.add_argument("--verbose", action="store_true", default=False,
                            help="Enable additional output.")

        parser.add_argument("--skip-broken", action="store_true", default=False,
                            help="Skip broken input files.  Without this things like duplicate "
                            "stanzas and invalid entries will cause processing to stop.")

        parser.add_argument("--match", "-m",  # metavar="MODE",
                            choices=["regex", "wildcard", "string"],
                            default="wildcard",
                            help=dedent("""\
            Specify pattern matching mode.
            Defaults to 'wildcard' allowing for ``*`` and  ``?`` matching.
            Use 'regex' for more power but watch out for shell escaping.
            Use 'string' to enable literal matching."""))
        parser.add_argument("--ignore-case", "-i", action="store_true",
                            help=dedent("""\
            Ignore case when comparing or matching strings.
            By default matches are case-sensitive."""))
        parser.add_argument("--invert-match", "-v", action="store_true",
                            help=dedent("""\
            Invert match results.
            This can be used to show what content does NOT match,
            or make a backup copy of excluded content."""))

        pg_out = parser.add_argument_group("Output mode", dedent("""\
            Select an alternate output mode.
            If any of the following options are used, the stanza output is not shown.
            """))
        pg_out.add_argument("--files-with-matches", "-l", action="store_true",
                            help="List files that match the given search criteria")
        pg_om1 = pg_out.add_mutually_exclusive_group()
        pg_om1.add_argument("--count", "-c", action="store_true",
                            help="Count matching stanzas")
        pg_om1.add_argument("--brief", "-b", action="store_true",
                            help="List name of matching stanzas")

        pg_sel = parser.add_argument_group("Stanza selection", dedent("""\
            Include or exclude entire stanzas using these filter options.

            All filter options can be provided multiple times.
            If you have a long list of filters, they can be saved in a file and referenced using
            the special ``file://`` prefix.  One entry per line.  Entries can be either a
            literal strings, wildcards, or regexes, depending on MATCH."""))

        pg_sel.add_argument("--stanza", metavar="PATTERN", action="append", default=[],
                            help=dedent("""
            Match any stanza who's name matches the given pattern.
            PATTERN supports bulk patterns via the ``file://`` prefix."""))

        pg_sel.add_argument("--attr-present", metavar="ATTR", action="append", default=[],
                            help=dedent("""\
            Match any stanza that includes the ATTR attribute.
            ATTR supports bulk attribute patterns via the ``file://`` prefix."""))

        pg_sel.add_argument("--attr-matches",
                            "--attr-eq",
                            metavar=("ATTR", "PATTERN"), nargs=2, action="append",
                            default=[],
                            help=dedent("""
            Match any stanza containing ATTR == PATTERN.
            PATTERN supports the special ``file://filename`` syntax.  Matching can be a direct
            string comparison (equals), or a regex and wildcard match.

            Note that all ``--attr-match`` and ``--attr-not-match`` arguments are matched together.
            For a stanza to match, all rules must apply.
            If attr is missing from a stanza, the value becomes an empty string for matching purposes."""))

        pg_sel.add_argument("--attr-not-matches",
                            "--attr-ne",
                            metavar=("ATTR", "PATTERN"), nargs=2, action="append",
                            default=[],
                            help=dedent("""
            Match any stanza containing ATTR != PATTERN.
            See ``--attr-matches`` for additional details."""))

        pg_eod = pg_sel.add_mutually_exclusive_group()
        pg_eod.add_argument("-e", "--enabled-only", action="store_true",
                            help=dedent("""
            Keep only enabled stanzas.  Any stanza containing ``disabled = 1`` will be removed.
            The value of ``disabled`` is assumed to be false by default.
            """))

        pg_eod.add_argument("-d", "--disabled-only", action="store_true",
                            help=dedent("""
            Keep disabled stanzas only.
            The value of the `disabled` attribute is interpreted as a boolean. """))

        pg_con = parser.add_argument_group("Attribute selection", dedent("""\
            Include or exclude attributes passed through.
            By default, all attributes are preserved.
            Allowlist (keep) operations are preformed before blocklist (reject) operations."""))

        pg_con.add_argument("--keep-attrs", metavar="WC-ATTR", default=[], action="append",
                            help=dedent("""\
            Select which attribute(s) will be preserved.
            This space separated list of attributes indicates what to preserve.
            Supports wildcards."""))

        pg_con.add_argument("--reject-attrs", metavar="WC-ATTR", default=[], action="append",
                            help=dedent("""\
            Select which attribute(s) will be discarded.
            This space separated list of attributes indicates what to discard.
            Supports wildcards."""))

    def prep_filters(self, args):
        flags = 0
        if args.ignore_case:
            flags |= FilteredList.IGNORECASE
        if args.verbose:
            flags |= FilteredList.VERBOSE

        self.stanza_filters = create_filtered_list(args.match, flags).feedall(args.stanza)
        self.attr_presence_filters = create_filtered_list(args.match, flags)
        self.attr_presence_filters.feedall(args.attr_present)

        # Q:  Should we check to see if the same attribute is used more than once (likely a typo?)
        # A:  No, let's trust the user; and avoid code bloat for hypothetical mistakes.
        self.attr_value_filters: List[Tuple[str, FilteredList]] = []
        if args.attr_matches:
            for attr, value in args.attr_matches:
                value_filter = create_filtered_list(args.match, flags)
                value_filter.feed(value)
                self.attr_value_filters.append((attr, value_filter))
        if args.attr_not_matches:
            for attr, value in args.attr_not_matches:
                value_filter = create_filtered_list(args.match, flags | FilteredList.INVERT)
                value_filter.feed(value)
                self.attr_value_filters.append((attr, value_filter))

        if args.enabled_only:
            self.disabled_filter = lambda attrs: not is_disabled(attrs)
        elif args.disabled_only:
            self.disabled_filter = is_disabled
        else:
            self.disabled_filter = lambda attrs: True

        if args.keep_attrs or args.reject_attrs:
            self.attrs_keep_filter = FilteredListWildcard(flags)
            for attrs in args.keep_attrs:
                self.attrs_keep_filter.feedall(attrs.split(" "))
            self.attrs_reject_filter = FilteredListWildcard(FilteredList.INVERT | flags)
            for attrs in args.reject_attrs:
                self.attrs_reject_filter.feedall(attrs.split(" "))
        else:
            # Bypass filter
            self.filter_attrs = lambda x: x

    def _test_stanza(self, stanza: str, attributes: dict) -> bool:
        if self.stanza_filters.match_stanza(stanza):
            # Exclude based on value of 'disabled' attribute
            if not self.disabled_filter(attributes):
                return False

            # If attr matching is in use, then test all attribute/match.  All must match.
            if self.attr_value_filters:
                for attr_name, attr_filter in self.attr_value_filters:
                    value = attributes.get(attr_name, "")
                    if not attr_filter.match(value):
                        return False
                return True

            # If there are no attribute level filters, automatically keep (preserves empty stanzas)
            if not self.attr_presence_filters.has_rules:
                return True
            # See if any of the attributes we are looking for exist, if so keep the entire stanza
            for attr in attributes:
                if self.attr_presence_filters.match(attr):
                    return True
        return False

    def filter_attrs(self, content: dict) -> dict:
        d = {}
        for (attr, value) in content.items():
            if self.attrs_keep_filter.match(attr) and self.attrs_reject_filter.match(attr):
                d[attr] = content[attr]
        return d

    def output(self, args, matches: dict, filename):
        """
        Process output for a single input file.
        """
        if args.files_with_matches:
            if matches:
                if args.count:
                    args.output.write(f"{filename} has {len(matches)} matching stanza(s)\n")
                elif args.brief:
                    for stanza_name in matches:
                        args.output.write(f"{filename}: {stanza_name}\n")
                else:
                    # Just show a single file
                    args.output.write(f"{filename}\n")
            elif args.verbose:
                self.stderr.write(f"No matching stanzas in {filename}\n")
        elif args.count:
            args.output.write(f"{len(matches)}\n")
        elif args.brief:
            for stanza_name in matches:
                args.output.write(f"{stanza_name}\n")
        else:
            if len(args.conf) > 1:
                args.output.write(f"#  {filename}\n")
            if matches:
                write_conf_stream(args.output, matches)
            elif args.verbose:
                self.stderr.write(f"No matching stanzas in {filename}\n")
            if args.verbose:
                sys.stderr.write(f"Matched {len(matches)} stanzas from {filename}\n")

    def run(self, args):
        ''' Filter configuration files. '''
        self.prep_filters(args)

        # By allowing multiple input CONF files, this means that we could have duplicate stanzas
        # (not detected by the parser)
        # so for now that just means duplicate stanzas on the output, but that may be problematic
        # I guess this is really up to the invoker to know if they care about that or not...
        # Still would be helpful for a quick "grep" of a large number of files

        for conf in args.conf:
            try:
                conf.set_parser_option(keep_comments=args.comments)
                cfg = conf.data
            except ConfParserException as e:
                action = "Aborting"
                if args.skip_broken:
                    action = "Skipping"
                self.stderr.write(f"{action} due to parsing error during {conf.name} due to {e}\n")
                if action == "Aborting":
                    return EXIT_CODE_BAD_CONF_FILE
                continue

            cfg_out = dict()
            for stanza_name, attributes in cfg.items():
                keep = self._test_stanza(stanza_name, attributes) ^ args.invert_match
                if keep:
                    cfg_out[stanza_name] = self.filter_attrs(attributes)

            self.output(args, cfg_out, conf.name)
            # Explicit flush used to resolve a CLI unittest timing issue in pypy
            args.output.flush()

        return EXIT_CODE_SUCCESS



# -*- coding: utf-8 -*-
"""
SUBCOMMAND:  ``ksconf rest-publish <ENDPOINT> <CONF>``

Usage example:

.. code-block:: sh

    ksconf rest-publish MyApp/local/props.conf


"""
from __future__ import absolute_import, unicode_literals

import os
import sys
from argparse import ArgumentParser, Namespace
from typing import Optional
from urllib.parse import urlparse

from ksconf.command import (ConfFileProxy, ConfFileType, KsconfCmd,
                            add_splunkd_access_args, add_splunkd_namespace,
                            dedent)
from ksconf.conf.delta import DiffHeader, compare_stanzas, is_equal, reduce_stanza, show_diff
from ksconf.conf.meta import MetaData
from ksconf.conf.parser import GLOBAL_STANZA, PARSECONF_LOOSE, ConfType, conf_attr_boolean
from ksconf.consts import EXIT_CODE_SUCCESS
from ksconf.util.completers import conf_files_completer

# Lazy loaded by _handle_imports()
splunklib = None


class RestPublishCmd(KsconfCmd):
    help = "Publish .conf settings to a live Splunk instance via REST"
    description = dedent("""\
    Publish stanzas in a .conf file to a running Splunk instance via REST.  This requires access to
    the HTTPS endpoint of Splunk.  By default, ksconf will handle both the creation of new stanzas
    and the update of existing stanzas.

    This can be used to push full configuration stanzas where you only have REST access and can't
    directly publish an app.

    Only attributes present in the conf file are pushed.  While this may seem obvious, this fact can
    have profound implications in certain situations, like when using this command for continuous
    updates.  This means that it's possible for the source .conf to ultimately differ from what ends
    up on the server's .conf file.  One way to avoid this, is to explicitly remove an object using
    ``--delete`` mode first, and then insert a new copy of the object.  Of course, this means that
    the object will be unavailable.  The other impact is that diffs only compares and shows a subset
    of attribute.

    Be aware, that for consistency, the configs/conf-TYPE endpoint is used for this command.
    Therefore, a reload may be required for the server to use the published config settings.
    """)

    maturity = "alpha"

    def __init__(self, *args, **kwargs):
        super(RestPublishCmd, self).__init__(*args, **kwargs)
        self._service = None
        self.meta: Optional[MetaData] = None

    @classmethod
    def _handle_imports(cls):
        g = globals()
        if globals()["splunklib"]:
            return
        import splunklib.client
        g["splunklib"] = splunklib
        import splunklib
        cls.version_extra = f"splunk-sdk {splunklib.__version__}"

    def register_args(self, parser: ArgumentParser):
        parser.add_argument("conf", metavar="CONF", nargs="+",
                            type=ConfFileType("r", "load", parse_profile=PARSECONF_LOOSE),
                            help="Configuration file(s) to export settings from."
                            ).completer = conf_files_completer

        parser.add_argument("--conf", dest="conf_type", metavar="TYPE",
                            help=dedent("""\
            Explicitly set the configuration file type.  By default, this is derived from CONF, but
            sometimes it's helpful to set this explicitly. Can be any valid Splunk conf file type.
            Examples include: 'app', 'props', 'tags', 'savedsearches', etc."""))
        parser.add_argument("-m", "--meta", action="append",
                            help="Specify one or more ``.meta`` files to determine the desired read & "
                            "write ACLs, owner, and sharing for objects in the CONF file.")

        # add_splunkd_namespace(
        #    add_splunkd_access_args(parser.add_argument("Splunkd endpoint")))

        add_splunkd_namespace(parser)
        add_splunkd_access_args(parser)

        parsg1 = parser.add_mutually_exclusive_group(required=False)
        '''
        parsg1.add_argument("-u", "--update", action="store_true", default=False,
                            help="Assume that the REST entities already exist.")
        parsg1.add_argument("--update-only", action="store_true", default=False,
                            help="Only update existing entities.  "
                                 "Non-existent entries will be skipped.")
        '''
        parsg1.add_argument("-D", "--delete", action="store_true", default=False,
                            help=dedent("""\
            Remove existing REST entities.  This is a destructive operation.
            In this mode, stanza attributes are unnecessary.
            NOTE:  This works for 'local' entities only; the default folder cannot be updated.
            """))

    @staticmethod
    def make_boolean(stanza: ConfType, attr: str = "disabled"):
        if attr in stanza:
            stanza[attr] = "1" if conf_attr_boolean(stanza[attr]) else "0"

    def connect_splunkd(self, args: Namespace):
        up = urlparse(args.url)
        # Take username/password form URL, if encoded there; otherwise use defaults from argparse
        if args.token:
            auth_args = {
                "splunkToken": args.token
            }
            login_fail_info = f"splunkToken={args.token[:10]}..."
        elif args.session_key:
            auth_args = {
                "token": args.session_key
            }
            login_fail_info = f"session={args.session_key[:10]}..."
        else:
            username = up.username or args.user
            password = up.password or args.password
            auth_args = {
                "username": username,
                "password": password,
            }
            login_fail_info = f"user={username} pass={'*' * len(password)}"
        try:
            self._service = splunklib.client.connect(
                host=up.hostname, port=up.port,
                owner=args.owner, app=args.app, sharing=args.sharing,
                verify=False if args.insecure else True,
                **auth_args)
            # Sanity check to:
            #   (1) confirm that session key is good, and
            #   (2) confirm that that the given namespace (app) is legit.
            self._service.apps.list()
        except Exception as e:
            sys.stderr.write(f"Connect issue url=https://{up.hostname}:{up.port} "
                             f"{login_fail_info}:  {e}\n")
            raise e

    def handle_conf_file(self, args: Namespace, conf_proxy: ConfFileProxy):
        if args.conf_type:
            conf_type = args.conf_type
        else:
            conf_type = os.path.basename(conf_proxy.name).replace(".conf", "")

        try:
            config_file = self._service.confs[conf_type]
        except KeyError:
            self.stderr.write(f"Invalid conf type named '{conf_type}'.\n")
            return
        conf = conf_proxy.data

        # Sorting stanza for consistent processing of large files.  No CLI option for now.
        # XXX:  Support stanza order preservation after new parser is created (long-term)
        for stanza_name in sorted(conf):
            stanza_data = conf[stanza_name]

            if not stanza_data:
                sys.stderr.write(f"Skipping empty stanza [{stanza_name}]\n")
                continue

            if stanza_name is GLOBAL_STANZA or stanza_name == "":
                # XXX:  Research proper handling of default/global stanzas..
                # As-is, curl returns an HTTP error, but yet the new entry is added to the
                # conf file.  So I suppose we could ignore the exit code?!    ¯\_(ツ)_/¯
                sys.stderr.write("Refusing to touch the [default] stanza.  Too much could go wrong.\n")
                continue

            if args.delete:
                action, info = self.delete_conf(stanza_name, stanza_data, config_file)
            else:
                action, info = self.publish_conf(stanza_name, stanza_data, config_file)

            self.stdout.write(f"{stanza_name:50} {action:8}   (delta size: { len(info.get('delta', [])) })\n")

            update_time = info.get("updated", 0)
            # headers = (conf_proxy.name, f"{args.url}/{config_file.path}")
            # rest_header = DiffHeader(f"{args.url}/{info.get('path', config_file.path)}", update_time)
            rest_header = DiffHeader(info.get("path", config_file.path), update_time)
            if action != "nochange" and "delta" in info:
                show_diff(self.stdout, info["delta"], headers=(conf_proxy.name, rest_header))

            if "meta" in info:
                self.stdout.write(f'{info["meta"]}\n')

            if "acl_delta" in info:
                show_diff(self.stdout, info["acl_delta"])

    def publish_conf(self,
                     stanza_name: str,
                     stanza_data: ConfType,
                     config_file):
        if self.meta:
            namespace = self._service.namespace
            metadata = self.meta.get(config_file.name, stanza_name)
            owner = metadata.get("owner", namespace.owner)
            app = namespace.app
            if metadata.get("export", None) == "system":
                sharing = "global"
            else:
                # Could still be "user" technically; but it seems unlikely that '--meta' would be given
                # in that case.  Still, there's possible room for improvement.
                sharing = namespace.sharing
        else:
            metadata = {}
            owner = None
            sharing = None
            app = None

        res = {}
        # XXX:  Move boolean comparison stuff to the core delta detection library....
        self.make_boolean(stanza_data)

        try:
            stz = config_file[stanza_name]
        except KeyError:
            stz = None

        if stz is not None:
            # print(f"Stanza {stanza_name} already exists on server.  Checking to see if update is needed.")
            # When pulling do we need to specify this?  (owner=owner, app=app, sharing=sharing);
            # If meta is given and where these are different than the defaults on the CLI?...
            stz_data = stz.content

            # Diff printing really doesn't like 'None's...
            stz_data = {k: v or "" for k, v in stz_data.items()}
            self.make_boolean(stz_data)
            res["path"] = stz.path
            try:
                res["updated"] = stz.state["updated"]
            except Exception:
                pass
            # print(f"VALUE NOW:   (FROM SERVER)   {stz.content}")  ## VERY NOISY!
            data = reduce_stanza(stz_data, stanza_data)
            # print(f"VALUE NOW:   (FILTERED TO OUR ATTRS)   {data}")
            delta = res["delta"] = compare_stanzas(stanza_data, data, stanza_name)
            if is_equal(delta):
                # print("NO CHANGE NEEDED.")
                res["delta"] = []
                action = "nochange"
            else:
                stz.update(**stanza_data)
                # Any need to call .refresh() here to grab the state from the server?
                action = "update"
        else:
            # print(f"Stanza {stanza_name} new -- publishing!")
            stz = config_file.create(stanza_name, owner=owner, app=app, sharing=sharing, **stanza_data)
            res["delta"] = compare_stanzas({}, stanza_data, stanza_name)
            res["path"] = stz.path
            action = "new"

        # METADATA PUSH

        if not self.meta:
            return (action, res)

        if not int(stz.access["can_change_perms"]):
            res["meta"] = "Can't change meta according to 'can_change_perms'"
            return (action, res)

        # NOTE:  We don't support attribute-level metadata here (Need it?  2 words:  pull request)
        if not metadata:
            res["meta"] = f"No metadata found for [{config_file.name}/{stanza_name}]"
            return (action, res)
        final_meta = {}
        if "access.read" in metadata:
            final_meta["perms.read"] = ",".join(metadata["access.read"])
        if "access.write" in metadata:
            final_meta["perms.write"] = ",".join(metadata["access.write"])
        if "owner" in metadata:
            final_meta["owner"] = metadata["owner"]
        else:
            final_meta["owner"] = "nobody"
        export = metadata.get("export", "")
        if export == "system":
            final_meta["sharing"] = "global"
        else:
            # Could still be "user" technically; but it seems unlikely that '--meta' would be given
            # in that case.  Still, there's possible room for improvement.
            final_meta["sharing"] = "app"

        # Build access dict for comparison purpose
        access = {}
        for x in ("owner", "app", "sharing"):
            access[x] = stz.access[x]
        for x in ("read", "write"):
            try:
                access["perms." + x] = ",".join(stz.access["perms"][x])
            except (KeyError, TypeError):
                access["perms." + x] = ""
        # print(f"[{stanza_name}] fm={final_meta} access:  {access}")

        acl_delta = compare_stanzas(reduce_stanza(access, final_meta), final_meta,
                                    stanza_name + "/acl")
        if is_equal(acl_delta):
            res["acl_delta"] = []
            return (action, res)
        else:
            res["acl_delta"] = acl_delta

        resource = None
        try:
            # Wonky workaround.  See https://github.com/splunk/splunk-sdk-python/issues/207
            # config_file.service.http.post()
            # response = Endpoint(config_file.service, stz.path + "acl/").post(**final_meta)

            svc = config_file.service
            all_headers = svc.additional_headers + svc._auth_headers
            prefix = stz.path
            if not prefix.endswith("/"):
                prefix += "/"

            resource = svc.authority + \
                svc._abspath(prefix + "acl",
                             owner=svc.namespace.owner, app=svc.namespace.app,
                             sharing=svc.namespace.sharing)
            # logger.debug("request to do the ACL THING!  (Round trip debugging)")
            response = svc.http.post(resource, all_headers, **final_meta)

            res["meta_response"] = response
        except Exception:
            # Don't die on exceptions for ACLs...  print the error and move on (too many things to go wrong here)
            sys.stderr.write(f"Failed hitting:  {resource}  ARGS={final_meta}\n")
            import traceback
            traceback.print_exc()
            # XXX:  Do better

        return (action, res)

    def delete_conf(self,
                    stanza_name: str,
                    stanza_data: ConfType,
                    config_file):
        res = {}
        if stanza_name in config_file:
            stz = config_file[stanza_name]
            stz_data = stz.content
            res["path"] = stz.path
            try:
                res["updated"] = stz.state["updated"]
            except KeyError:
                # Doesn't matter
                pass

            self.make_boolean(stz_data)
            # print(f"Found {stz_data}")
            data = reduce_stanza(stz_data, stanza_data)
            config_file.delete(stanza_name)
            res["delta"] = compare_stanzas(data, {}, stanza_name)
            return ("deleted", res)
        else:
            res["delta"] = []
            return ("nochange", res)

    def run(self, args: Namespace):
        if args.meta:
            self.meta = MetaData()
            for meta_file in args.meta:
                self.stdout.write(f"Loading metadata from {meta_file}\n")
                self.meta.feed_file(meta_file)

        self.connect_splunkd(args)
        for conf_proxy in args.conf:    # type: ConfFileProxy
            self.handle_conf_file(args, conf_proxy)

        return EXIT_CODE_SUCCESS


'''
# For really crapy debug messages from splunksdk  (optimize server round trips)

import logging

logging.basicConfig()
logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
'''



""" SUBCOMMAND:  ``ksconf xml-format <XML>``

Usage example:   (Nice pre-commit script)

.. code-block:: sh

    find default/data/ui -name '*.xml' | ksconf xml-format -

"""
from __future__ import absolute_import, unicode_literals

import os
from argparse import SUPPRESS
from collections import Counter

from ksconf.command import KsconfCmd, dedent
from ksconf.consts import (EXIT_CODE_BAD_CONF_FILE, EXIT_CODE_FORMAT_APPLIED,
                           EXIT_CODE_INTERNAL_ERROR, EXIT_CODE_SUCCESS)
from ksconf.util import debug_traceback
from ksconf.util.completers import conf_files_completer
from ksconf.util.file import _stdin_iter
# Main codebase is here:
from ksconf.xmlformat import SplunkSimpleXmlFormatter

# Lazy loaded by _handle_imports()
etree = None


class XmlFormatCmd(KsconfCmd):
    help = "Normalize XML view and nav files"
    description = dedent("""
    Normalize and apply consistent XML indentation and CDATA usage for XML dashboards and
    navigation files.

    Technically this could be used on *any* XML file, but certain element names specific to Splunk's
    simple XML dashboards are handled specially, and therefore could result in unusable results.

    The expected indentation level is guessed based on the first element indentation, but can be
    explicitly set if not detectable.
    """)
    maturity = "alpha"

    @classmethod
    def _handle_imports(cls):
        g = globals()
        if globals()["etree"]:
            return
        from lxml import etree
        cls.version_extra = f"lxml {etree.__version__}"
        g["etree"] = etree

    def register_args(self, parser):
        parser.add_argument("xml", metavar="FILE", nargs="+", help=dedent("""\
            One or more XML files to check.
            If '-' is given, then a list of files is read from standard input""")
                            ).completer = conf_files_completer
        parser.add_argument("--indent", type=int, default=2,
                            help="Number of spaces.  This is only used if indentation cannot be "
                            "guessed from the existing file.")
        parser.add_argument("--quiet", "-q", default=False, action="store_true",
                            help="Reduce the volume of output.")

        # Hidden arguments
        parser.add_argument("--disable-pre-commit-migration-check",
                            default=False, action="store_true", help=SUPPRESS)

    def pre_commit_repo_migration_warning(self, args):
        r"""
        Issue migration warning if (1) running hooks from the old repo (missing
        arg), and (2) parent process is from pre-commit (env var).


        Another workaround is to use:

        ..  code-block:: yaml

            - repo: https://github.com/Kintyre/ksconf
            rev: v0.11.8
            hooks:
                - id: ksconf-check
                - id: ksconf-sort
                exclude: logging\.conf
                - id: ksconf-xml-format
                  args: --disable-pre-commit-migration-check
            additional_dependencies: [lxml]

        But honestly, isn't it just easy to add ``-pre-commit`` to the repo?

        Remove this after Dec 2024 or v0.13.0
        """
        # New repo uses the following config:
        # entry: ksconf xml-format -q --disable-pre-commit-migration-check
        # If this flag has been used, assume we're running from the correct repo
        if args.disable_pre_commit_migration_check:
            return

        # See if pre-commit is my parent.  Assume this based on env variable.
        if os.environ.get("PRE_COMMIT", "") != "1":
            return

        from warnings import warn
        warn("You appear to be using the 'ksconf-xml-format' pre-commit hook from the ksconf repo. "
             "The ksconf pre-commit hooks have been moved to a new repo.  "
             "This configuration will stop working after v0.13.0 "
             "Please update '.pre-commit-config.yaml' to use the new "
             "repo: https://github.com/Kintyre/ksconf-pre-commit.git")

    def run(self, args):
        self.pre_commit_repo_migration_warning(args)
        formatter = SplunkSimpleXmlFormatter()
        # Should we read a list of conf files from STDIN?
        if len(args.xml) == 1 and args.xml[0] == "-":
            files = _stdin_iter()
        else:
            files = args.xml
        c = Counter()

        # Copy this locally, to avoid an even stranger exception later
        ParseError = etree.ParseError

        exit_code = EXIT_CODE_SUCCESS
        for fn in files:
            c["checked"] += 1
            if not os.path.isfile(fn):
                self.stderr.write(f"Skipping missing file:  {fn}\n")
                c["missing"] += 1
                continue
            try:
                if formatter.format_xml(fn, fn, args.indent):
                    self.stderr.write(f"Replaced file {fn} with formatted content\n")
                    c["changed"] += 1
                else:
                    if not args.quiet:
                        self.stderr.write(f"Already formatted {fn}\n")
                    c["no-action"] += 1
                self.stderr.flush()
            except ParseError as e:
                self.stderr.write(f"Error parsing file {fn}:  {e}\n")
                self.stderr.flush()
                c["error"] += 1
                exit_code = EXIT_CODE_BAD_CONF_FILE
            except Exception as e:  # pragma: no cover
                self.stderr.write(f"Unhandled top-level exception while parsing {fn}.  "
                                  f"Aborting.\n{e}\n")
                debug_traceback()
                c["error"] += 1
                exit_code = EXIT_CODE_INTERNAL_ERROR
                break

        if not exit_code and c["changed"] > 0:
            exit_code = EXIT_CODE_FORMAT_APPLIED

        if True:  # show stats or verbose
            self.stdout.write(f"Completed formatting {c['checked']} files.  "
                              f"rc={exit_code} Breakdown:\n"
                              f"   {c['changed']} files were formatted successfully.\n"
                              f"   {c['no-action']} files were already formatted.\n"
                              f"   {c['error']} files failed.\n")
        return exit_code
